/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_ccf.h 161769 2014-05-30 14:32:34Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-30 16:32:34 +0200 (Fri, 30 May 2014) $
 * $Revision: 161769 $
 */

#ifndef ESPDA_CCF_H
#define ESPDA_CCF_H

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>
#include <string.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_min.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_statistics_double.h>

/*-----------------------------------------------------------------------------
                                   Defines
 -----------------------------------------------------------------------------*/

#define GAUSS_NPAR 4

/* colors */
#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_YELLOW  "\x1b[33m"
#define ANSI_COLOR_BLUE    "\x1b[34m"
#define ANSI_COLOR_MAGENTA "\x1b[35m"
#define ANSI_COLOR_CYAN    "\x1b[36m"
#define ANSI_COLOR_RESET   "\x1b[0m"


#define espda_msg(...) cpl_msg_info(cpl_func, __VA_ARGS__)
#define espda_msg_warning(...) cpl_msg_warning(cpl_func, __VA_ARGS__)
#define espda_msg_error(...) cpl_msg_error(__func__, __VA_ARGS__)

#define espda_ensure(CONDITION, ERROR_CODE, ...) \
    if (CONDITION) { \
        espda_msg_error(__VA_ARGS__); \
        cpl_ensure(0, ERROR_CODE, ERROR_CODE); }




/*----------------------------------------------------------------------------
 Data structures
 ----------------------------------------------------------------------------*/
typedef struct
{
    double *x, *y, *err, *fit;
    int n, m;
    double delta;
} espda_ngauss_data;

typedef struct
{
    double c, sig_c, *k, *sig_k, *x0, *sig_x0, *fwhm, *sig_fwhm;
    int n, m;
} espda_ngauss_result;

typedef struct
{
    double *x, *y, *err;
    int n;
} espda_poly_data;

/*-----------------------------------------------------------------------------
                                   Functions prototypes
 -----------------------------------------------------------------------------*/

cpl_error_code espda_compute_CCF(cpl_image *ll,
                                 cpl_image *dll,
                                 cpl_image *flux,
                                 cpl_image *error,
                                 cpl_image *blaze,
                                 cpl_image *quality,
                                 cpl_array *RV_table,
                                 cpl_table *mask,
                                 double mask_width,
                                 double berv,
                                 double bervmax,
                                 cpl_image **CCF_flux_RE,
                                 cpl_image **CCF_error_RE,
                                 cpl_image **CCF_quality_RE);

cpl_error_code espda_fit_ccf(cpl_array *RV_table,
                             cpl_image *CCF_flux,
                             cpl_image *CCF_error,
                             cpl_image *CCF_quality,
                             double *asym_RE,
                             double *asym_err_RE,
                             double *bispan_RE,
                             double *bispan_err_RE,
                             cpl_table **residuals_RE,
                             espda_ngauss_result *g_res_RE);

const char *espda_get_mask_name(const char *spec_type_id,
                                const cpl_table *mask_lut);

cpl_error_code espda_compute_rv_table(double RV_range,
                                      double RV_step,
                                      double RV_center,
                                      cpl_array **RV_table_RE);

cpl_error_code espda_fit_Ngauss (espda_ngauss_data *data,
                                 double *x0init,
                                 espda_ngauss_result *res,
                                 int spectrum_type,
                                 int print_flag);

cpl_error_code espda_fit_poly (espda_poly_data *data,
                               int deg,
                               double *fit_RE,
                               double *coeffs_RE,
                               double *coeffs_err_RE,
                               double *chisq_RE,
                               int print_flag);

cpl_error_code espda_remove_cosmics_via_histogram(cpl_image *flux,
                                                  cpl_image *err,
                                                  cpl_image *flux_RE,
                                                  int *cosmics_RE);

cpl_error_code espda_correct_flux(cpl_image *s2d_blaze,
                                  cpl_image *wave_matrix,
                                  cpl_table *flux_template,
                                  cpl_table *mask_table,
                                  char *spec_type,
                                  cpl_image *s2d_blaze_flux_corr_RE,
                                  double *flux_corr);

cpl_error_code espda_dfs_save_data_err_qual(cpl_frameset *all_frames,
                                            const cpl_parameterlist *parameters,
                                            const cpl_frameset *used_frames,
                                            const char *recipe,
                                            cpl_propertylist *keywords,
                                            const char *filename,
                                            cpl_image **images_to_save);

cpl_error_code espda_compu_radvel_QC(cpl_array *RV_table,
                                     double RV_step,
                                     int orders_nb,
                                     espda_ngauss_result *g_res,
                                     const char *ccf_mask_id,
                                     int *cosmics_nb_2,
                                     double *flux_correction,
                                     double asym,
                                     double asym_err,
                                     double bispan,
                                     double bispan_err,
                                     cpl_propertylist **keywords_RE);
#endif
