/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_model.c 161538 2014-05-22 13:56:10Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-22 15:56:10 +0200 (Thu, 22 May 2014) $
 * $Revision: 161538 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_model.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup  espda_model  Modelization
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Model the column density of a Lyman-alpha line.
  @param   peak  Peak value of the density flux [erg cm^-2 s^-1 nm^-1].
  @param   cont  Continuum values at peak [erg cm^-2 s^-1 nm^-1].
  @return  The approximate column density [cm^-2, logarithm].
   
  This function estimates the column density of a Ly-alpha line from the 
  continuum-normalized flux at line peak @a peak/@a cont.
    
  The column density is computed from a degree-5 polynomial approximation of 
  the curve of growth of HI, assuming that the Doppler broadening is 20 km/s:
  @f[
    \log(N_\mathrm{HI})\simeq 14.096    - 4.6251f   + 18.657f^2 
                            - 46.299f^3 + 53.301f^4 - 23.442f^5
  @f]
  with @f$ f=\mathit{peak}/\mathit{cont} @f$.

  N.B. The function is provisional. Better modelizatiosn will be implemented in    
       the future.
 */
/*----------------------------------------------------------------------------*/

double espda_model_colden(const double peak, const double cont) {
    double a_0 	  =  14.096;
    double a_1 	  = - 4.6251;
    double a_2 	  =  18.657;
    double a_3 	  = -46.299;
    double a_4 	  =  53.301;
    double a_5 	  = -23.442;
    double colden = a_0 + a_1 *    (peak / cont)    +
                        + a_2 * pow(peak / cont, 2) +
                        + a_3 * pow(peak / cont, 3) +
                        + a_4 * pow(peak / cont, 4) +
                        + a_5 * pow(peak / cont, 5);
    return colden;
}

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

/*
double espda_model_colden2(const char *line_id, const double peak) {
    FILE *growth_file = fopen("/Users/guido/Work/espdap/espda/tests/GROWTH.fits", 
        "r");
    cpl_table *growth = NULL;
    if (growth_file) {
        growth = cpl_table_load("/Users/guido/Work/espdap/espda/tests/GROWTH.fits", 1, 0);
        fclose(growth_file);
    }
    
    cpl_table_subtract_scalar(growth, line_id, peak);
    cpl_table_abs_column(growth, line_id);

    int    x_a = cpl_table_get_column_minpos(growth, line_id);
    double v_a = cpl_table_get_column_min(growth, line_id);
    
    if (cpl_table_get_double(growth, line_id, ))
    int j_a = fmin(fmax((int)floor(j_0), 0), j_max - 2);
    int j_b = fmin(fmax((int)ceil (j_0), 1), j_max - 1);
    double v_aa  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                   "VOIGT_FUNC", j_a), k_a, NULL);
    double v_ba  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                   "VOIGT_FUNC", j_b), k_a, NULL);
            double v_ab  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                           "VOIGT_FUNC", j_a), k_b, NULL);
            double v_bb  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                           "VOIGT_FUNC", j_b), k_b, NULL);
    double colden 

    return colden;
}
*/

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

/*
int espda_model_costf(const gsl_vector *par_vary,
                      void             *data,
                      gsl_vector       *f) {

    cpl_array *fit_conv = NULL;
    gsl_matrix *jac = NULL; 
    espda_model_transm(par_vary, data, &fit_conv, &jac);
    //cpl_array_dump(fit_conv, 0, 10, NULL);
    cpl_table *spec_chunk = ((struct gsl_multifit_data *)data)->spec_chunk;
    size_t     size_chunk = (int)cpl_table_get_nrow(spec_chunk);
    for (int i = 0; i < (int)size_chunk; i++) {
        double fiti     = cpl_array_get_double(fit_conv, i, NULL);
	//printf("%f\n", fiti);
        double normi    = cpl_table_get_double(spec_chunk, "FLUX",    i, NULL) /
                          cpl_table_get_double(spec_chunk, "CONT",    i, NULL);
        double normerri = cpl_table_get_double(spec_chunk, "FLUXERR", i, NULL) /
                          cpl_table_get_double(spec_chunk, "CONT",    i, NULL);
        gsl_vector_set(f, i, (fiti - normi) / normerri);
    }
    return GSL_SUCCESS;
}
*/

/*----------------------------------------------------------------------------*/
/**
 */
/*----------------------------------------------------------------------------*/

/*
int espda_model_costdf(const gsl_vector *par_vary,
                       void             *data,
                       gsl_matrix       *J) {
//int espda_model_costdf(CBLAS_TRANSPOSE_t TransJ, const gsl_vector * x,
//            const gsl_vector * u, void * params, gsl_vector * v,
//            gsl_matrix * JTJ) {

    cpl_array *fit_conv = NULL;
    espda_model_transm(par_vary, data, &fit_conv, &J);
    //gsl_matrix_fprintf(stdout, J, "%e");
    return GSL_SUCCESS;
}
*/

/*----------------------------------------------------------------------------*/
/**
  @brief   Model the instrument profile.
  @param   size   The size of the model.
  @param   resol  The resolution of model.
  @param   kappa  Number of FWHM used to truncate the gaussian.
  @return  The instrument profile (INST_PROF).
   
  This function models the instrument profile as a Gaussian with size @a size, 
  resolution @a resol, and peak equal to one. It is saved in column @a PROF 
  of the output table. The Gaussian is centered on the median row of the table 
  and its FWHM is @a size / @a kappa. Columns @a VEL and @a VELRED are also 
  computed.

  The Gaussian profile is defined at the @f$ i @f$-th pixel as @f[
    G[\mathit{i}]=\mathrm{exp}\left\{-\frac{1}{2}\left[
                  \frac{i-1/2(\mathit{size}-1)}{\mathit{size}}\times
                  \frac{\mathit{kappa}\times\mathrm{FWHM}}{\sigma}
                  \right]^2\right\}
  @f]
  with @f$ i @f$ ranging from 0 to @f$ size - 1 @f$. Here @f$ \sigma @f$ is the 
  standard deviation and @f$ \mathrm{FWHM} = c/\mathit{resol} @f$ is the 
  full-width at half-maximum, @f$ c @f$ being the speed of light. 
  @f$ \mathrm{FWHM}/\sigma @f$ is fixed for a Gaussian profile and is 
  approximately equal to 2.345820045. @a VEL and @a VELRED are computed as 
  follows: @f[
    \mathit{VEL}=\frac{i-1/2(\mathit{size}-1)}{\mathit{size}}\times
                 \mathit{kappa}\times\mathrm{FWHM},
  @f]
  @f[
    \mathit{VELRED}=\frac{i-1/2(\mathit{size}-1)}{\mathit{size}}\times
                 \frac{\mathit{kappa}}{\mathit{resol}}.
  @f]
  

  N.B. The function is provisional. It will be adapted to the actual instrument
       profile when it is available.
 */
/*----------------------------------------------------------------------------*/

cpl_table *espda_model_inst(cpl_size size, double resol, double kappa) {
    cpl_table *inst_prof = cpl_table_new(size);
    cpl_table_new_column(inst_prof, "VEL",    CPL_TYPE_DOUBLE);
    cpl_table_new_column(inst_prof, "VELRED", CPL_TYPE_DOUBLE);
    cpl_table_new_column(inst_prof, "PROF",   CPL_TYPE_DOUBLE);
    for (int i = 0; i < size; i++) {

        // Gaussian profile (FWHM = c / resolution; width = kappa * sigma)
        double vel   = ((double)i - (size - 1.0) * 0.5) / size *
                       LIGHTSPEED / resol * kappa;
        double velred = ((double)i - (size - 1.0) * 0.5) / size * kappa / resol;
        double prof  = exp(-0.5 * 
                       pow(((double)i - (size - 1.0) * 0.5) / size *
                       FWHM_TO_STDEV * kappa, 2));
        cpl_table_set_double(inst_prof, "VEL",  i, vel);
        cpl_table_set_double(inst_prof, "VELRED",  i, velred);
        cpl_table_set_double(inst_prof, "PROF", i, prof);
    }
    return inst_prof;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Model the transmission profile for a line list
  @param    par_vary  Varying parameters
  @param    data      Spectrum and ancillary data, including fixed parameters 
  @param    model     Transmission model
  @return   GSL_SUCCESS iff OK

  This function creates a model of the transmitted (normalized) flux for a list
  of absorption lines. The model is obtained by computing the Voigt profiles of
  all lines and summing them into a single optical depth term. The resulting 
  transmission profile is then convoluted with the instrument profile. To avoid
  border effects, the transmission profile is initially computed in a portion of
  spectrum ("over-chunk") slightly larger than the one obtained after the
  convolution ("chunk").

  The function is suitable to be handled by gsl_multifit. Varying line
  parameters must be fed separately as a GSL vector for the fitting procedure 
  to work properly. All the remaining information, including the tables with 
  the spectral chunk and over-chunk, the reference list of ionic transition, 
  and the fixed line parameters, must be provided separately within the 
  structure data.

  NOT COMPLETED! USE espda_model_transm_old
 */
/*----------------------------------------------------------------------------*/

/*
cpl_array *espda_model_transm(const gsl_vector  *par_vary,
                              void              *data,
			      cpl_array        **fit_conv,
                              gsl_matrix       **jac) 

{
    // Sort chunk and overchunks by ascending wavelength
    cpl_table *spec_over  = ((struct gsl_multifit_data *)data)->spec_over;  
    cpl_table *spec_chunk = ((struct gsl_multifit_data *)data)->spec_chunk;
    //cpl_table *lfit_chunk = ((struct gsl_multifit_data *)data)->lfit_chunk;
    size_t     size_over  = (int)cpl_table_get_nrow(spec_over);
    size_t     size_chunk = (int)cpl_table_get_nrow(spec_chunk);
    cpl_propertylist *spec_col    = cpl_propertylist_new();
    cpl_propertylist_append_bool(spec_col, "WAVEL", FALSE);
    cpl_table_sort(spec_over,  spec_col);
    cpl_table_sort(spec_chunk, spec_col);
    cpl_propertylist_delete(spec_col);

    // Prepare arrays for fitting
    cpl_array *wavel_over  = cpl_array_new(size_over,  CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double(wavel_over, 
        cpl_table_get_data_double(spec_over, "WAVEL")); 
    cpl_array *wavel_chunk  = cpl_array_new(size_chunk,  CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double(wavel_chunk, 
        cpl_table_get_data_double(spec_chunk, "WAVEL")); 
    cpl_array *tot = cpl_array_new(size_over,  CPL_TYPE_DOUBLE);
    cpl_array_fill_window_double(tot,   0, size_over,  0); 
    cpl_array *voigt = NULL, *fit_part   = NULL;

    // Merge Voigt parameters
    cpl_array *par_all = NULL;
    espda_voigt_merge(NULL, 
                    ((struct gsl_multifit_data *)data)->ion_wavel, 
                    ((struct gsl_multifit_data *)data)->par_pos,
                    ((struct gsl_multifit_data *)data)->par_fix, par_vary, NULL,
                    ((struct gsl_multifit_data *)data)->par_lim, &par_all, NULL);

    // Loop over lines
    cpl_size   par_num = ((struct gsl_multifit_data *)data)->par_num;
    *jac = gsl_matrix_alloc(size_chunk, par_vary->size);
    //gsl_matrix_set_zero(*jac);
    //printf("%i %i\n", (*jac)->size1, (*jac)->size2);
    //printf("%i %i\n", size_chunk, par_vary->size);

    //printf("%i %i\n", size_chunk, par_num);

    gsl_function func;
    //double deriv_linez, deriv_colden, deriv_therb, deriv_turbb; 
    //double abserr_linez, abserr_colden, abserr_therb, abserr_turbb; 
    func.function = &espda_model_voigtf;
    for (int i = 0; i < (int)par_num / 4; i++) { 
        //Used for derivatives, now disabled 
        //int j_linez = cpl_array_get_int(
	//		  ((struct gsl_multifit_data *)data)->par_pos,
	//		  4 * i + 0, NULL);
        //int j_colden = cpl_array_get_int(
	//		  ((struct gsl_multifit_data *)data)->par_pos,
	//		  4 * i + 1, NULL);
        //int j_therb = cpl_array_get_int(
	//		  ((struct gsl_multifit_data *)data)->par_pos,
	//		  4 * i + 2, NULL);
        //int j_turbb = cpl_array_get_int(
	//		  ((struct gsl_multifit_data *)data)->par_pos,
	//		  4 * i + 3, NULL);

        // Compute the Voigt profile on over-chunks
        double ion_wavel = cpl_array_get_double(((struct gsl_multifit_data *)
                                                data)->ion_wavel, i, NULL);
        double ion_oscs  = cpl_array_get_double(((struct gsl_multifit_data *)
                                                data)->ion_oscs,  i, NULL);
        double ion_damp  = cpl_array_get_double(((struct gsl_multifit_data *)
                                                data)->ion_damp,  i, NULL);
        double linez  	 = cpl_array_get_double(par_all, 4 * i + 0, NULL);
        double colden 	 = cpl_array_get_double(par_all, 4 * i + 1, NULL);
        double therb  	 = cpl_array_get_double(par_all, 4 * i + 2, NULL);
        double turbb  	 = cpl_array_get_double(par_all, 4 * i + 3, NULL);
        //double peak;

        //espda_model_voigt_old(wavel_over,   ion_wavel, ion_oscs, 
        //              ion_damp, linez, colden, therb, turbb, 
        //               ((struct gsl_multifit_data *)data)->voigt_func,
        //                &voigt,   &fit_part, &peak);
	//struct model_voigt_params params = {ion_wavel, ion_oscs, 
        //                ion_damp, linez, colden, therb, turbb, 
        //                ((struct gsl_multifit_data *)data)->voigt_func};
	//struct model_voigtf_params params; #= {"none",
	//		   cpl_array_get_double(wavel_over, l, NULL), 
	//		   ion_wavel, ion_oscs, ion_damp, 
	//		   0.0, colden, therb, turbb, 
	//		   ((struct gsl_multifit_data *)data)->voigt_func};
	voigt = cpl_array_new(size_over, CPL_TYPE_DOUBLE);
	for (int l = 0; l < (int)size_over; l++) {
	    double voigt_value = espda_model_voigt(
		cpl_array_get_double(wavel_over, l, NULL), ion_wavel, ion_oscs, 
                        ion_damp, linez, colden, therb, turbb, 
                        ((struct gsl_multifit_data *)data)->voigt_func);
	    cpl_array_set_double(voigt, l, voigt_value);

	    //printf("%f %e\n", cpl_array_get_double(wavel_over, l, NULL), voigt_value);
	}    
	for (int l = 0; l < (int)size_chunk; l++) {

	    // Derivative w.r.t. LINEZ
	    struct model_voigtf_params params_linez = {"linez", 
		cpl_array_get_double(wavel_chunk, l, NULL), 
		ion_wavel, ion_oscs, ion_damp, 
		0.0, colden, therb, turbb, 
		((struct gsl_multifit_data *)data)->voigt_func};	
	    if (espda_model_voigtf(linez, &params_linez) > 0.0) {

	    	func.params = &params_linez;
	    	gsl_deriv_central(&func, linez, 1e-8, &deriv_linez, &abserr_linez); 
	    	gsl_matrix_set(*jac, l, j_linez, deriv_linez);
	    	//printf("%f %f\n", deriv_linez, abserr_linez);

	    	// Derivative w.r.t. COLDEN
	    	struct model_voigtf_params params_colden = {"colden", 
	    		cpl_array_get_double(wavel_chunk, l, NULL), 
	    		ion_wavel, ion_oscs, ion_damp, 
	    		0.0, colden, therb, turbb, 
	    		((struct gsl_multifit_data *)data)->voigt_func};	
	    	func.params = &params_colden;
	    	gsl_deriv_central(&func, colden, 1e-8, &deriv_colden, 
	    			      &abserr_colden); 
	    	gsl_matrix_set(*jac, l, j_colden, deriv_colden);

		//printf("%f %e %e\n", cpl_array_get_double(wavel_chunk, l, NULL),
		//      espda_model_voigtf(linez, &params_linez), deriv_linez);

	    	// Derivative w.r.t. THERB
	    	struct model_voigtf_params params_therb = {"therb", 
	    		cpl_array_get_double(wavel_chunk, l, NULL), 
	    		ion_wavel, ion_oscs, ion_damp, 
	    		0.0, colden, therb, turbb, 
	    		((struct gsl_multifit_data *)data)->voigt_func};	
	    	func.params = &params_therb;
	    	gsl_deriv_central(&func, therb, 1e-8, &deriv_therb, &abserr_therb); 
	    	gsl_matrix_set(*jac, l, j_therb, deriv_therb);

	    	// Derivative w.r.t. TURBB
	    	struct model_voigtf_params params_turbb = {"turbb", 
	    		cpl_array_get_double(wavel_chunk, l, NULL), 
	    		ion_wavel, ion_oscs, ion_damp, 
	    		0.0, colden, therb, turbb, 
	    		((struct gsl_multifit_data *)data)->voigt_func};	
	    	func.params = &params_turbb;
	    	gsl_deriv_central(&func, turbb, 1e-2, &deriv_turbb, &abserr_turbb); 
	    	gsl_matrix_set(*jac, l, j_turbb, deriv_turbb);
	    }
	}

        cpl_array_add(tot, voigt);
        cpl_array_delete(voigt);
        cpl_array_delete(fit_part);
    }
    cpl_array_delete(par_all);

    //gsl_matrix_fprintf(stdout, *jac, "%e");

    // Compute and save the overall fitting profile on over-chunks
    cpl_array *fit   = cpl_array_new(size_over, CPL_TYPE_DOUBLE);
    for (int l = 0; l < (int)size_over; l++) {
        cpl_array_set_double(fit, l, exp(-1.0 * 
                             cpl_array_get_double(tot, l, NULL)));
    }
    cpl_array_delete(tot);

    // Convolve the overall profile with the instrument profile on chunks
    *fit_conv = espda_spec_conv(((struct gsl_multifit_data *)data)
        ->prof_chunk, wavel_over, fit);
    cpl_array_delete(wavel_over);
    cpl_array_delete(fit); 
    cpl_table_copy_data_double(spec_chunk, "FIT", 
                               cpl_array_get_data_double(*fit_conv));
    
    //for (int i = 0; i < (int)par_num / 4; i++) { 
    //    
    //    cpl_table_set_double(lfit_chunk, "PEAK", i, );
    //}

    return *fit_conv;
}
*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Model the transmission profile for a line list
  @param    par_vary  Varying parameters
  @param    data      Spectrum and ancillary data, including fixed parameters 
  @param    model     Transmission model
  @return   GSL_SUCCESS iff OK

  This function creates a model of the transmitted (normalized) flux for a list
  of absorption lines. The model is obtained by computing the Voigt profiles of
  all lines and summing them into a single optical depth term. The resulting 
  transmission profile is then convoluted with the instrument profile. To avoid
  border effects, the transmission profile is initially computed in a portion of
  spectrum ("over-chunk") slightly larger than the one obtained after the
  convolution ("chunk").

  The function is suitable to be handled by gsl_multifit. Varying line
  parameters must be fed separately as a GSL vector for the fitting procedure 
  to work properly. All the remaining information, including the tables with 
  the spectral chunk and over-chunk, the reference list of ionic transition, 
  and the fixed line parameters, must be provided separately within the 
  structure data. 
 */
/*----------------------------------------------------------------------------*/

int espda_model_transm_old(const gsl_vector *par_vary,
                       void             *data,
                       gsl_vector       *model) 

{
    // Sort chunk and overchunks by ascending wavelength
  cpl_table *spec_over  = ((struct gsl_multifit_data *)data)->spec_over;  
    //#######
    cpl_table *spec_chunk = ((struct gsl_multifit_data *)data)->spec_chunk;
    /*
      double *spec_chunk = ((struct gsl_multifit_data *)data)->spec_chunk;
      cpl_array *spec_norm = ((struct gsl_multifit_data *)data)->spec_norm;
      cpl_array *spec_normerr = ((struct gsl_multifit_data *)data)->spec_normerr;
    */
    //#######

    //cpl_table *lfit_chunk = ((struct gsl_multifit_data *)data)->lfit_chunk;
    size_t     size_over  = (int)cpl_table_get_nrow(spec_over);

    //#######
    size_t     size_chunk = (int)cpl_table_get_nrow(spec_chunk);
    //#######

    cpl_propertylist *spec_col    = cpl_propertylist_new();
    cpl_propertylist_append_bool(spec_col, "WAVEL", FALSE);
    cpl_table_sort(spec_over,  spec_col);

    //#######
    cpl_table_sort(spec_chunk, spec_col);
    //#######

    cpl_propertylist_delete(spec_col);

    // Prepare arrays for fitting
    cpl_array *wavel_over  = cpl_array_new(size_over,  CPL_TYPE_DOUBLE);
    cpl_array_copy_data_double(wavel_over, 
        cpl_table_get_data_double(spec_over, "WAVEL")); 
    cpl_array *tot = cpl_array_new(size_over,  CPL_TYPE_DOUBLE);
    cpl_array_fill_window_double(tot,   0, size_over,  0); 
    cpl_array *voigt = NULL, *fit_part   = NULL;

    // Merge Voigt parameters
    cpl_array *par_all = NULL;
    cpl_ensure_code(espda_voigt_merge(NULL, 
                    ((struct gsl_multifit_data *)data)->ion_wavel, 
                    ((struct gsl_multifit_data *)data)->par_pos,
                    ((struct gsl_multifit_data *)data)->par_fix, par_vary, NULL,
                    ((struct gsl_multifit_data *)data)->par_lim, &par_all, NULL)
                    == CPL_ERROR_NONE, cpl_error_get_code());

    // Loop over lines
    cpl_size   par_num = ((struct gsl_multifit_data *)data)->par_num;
    for (int i = 0; i < (int)par_num / 4; i++) { 

        // Compute the Voigt profile on over-chunks
        double ion_wavel = cpl_array_get_double(((struct gsl_multifit_data *)
                                                data)->ion_wavel, i, NULL);
        double ion_oscs  = cpl_array_get_double(((struct gsl_multifit_data *)
                                                data)->ion_oscs,  i, NULL);
        double ion_damp  = cpl_array_get_double(((struct gsl_multifit_data *)
                                                data)->ion_damp,  i, NULL);
        double linez  	 = cpl_array_get_double(par_all, 4 * i + 0, NULL);
        double colden 	 = cpl_array_get_double(par_all, 4 * i + 1, NULL);
        double therb  	 = cpl_array_get_double(par_all, 4 * i + 2, NULL);
        double turbb  	 = cpl_array_get_double(par_all, 4 * i + 3, NULL);
        double peak;
        cpl_ensure_code(espda_model_voigt_old(wavel_over,   ion_wavel, ion_oscs, 
                        ion_damp, linez, colden, therb, turbb, 
                        ((struct gsl_multifit_data *)data)->voigt_func,
                        &voigt,   &fit_part, &peak) == CPL_ERROR_NONE, 
                        cpl_error_get_code());

        // If lines are more than one, sum individual profiles of all lines;
        // otherwise take take the fitting profile
        if ((int)par_num > 4) {
            cpl_array_add(tot, voigt);
        }
        else {
            cpl_array_add(tot, fit_part);
        }
        cpl_array_delete(voigt);
        cpl_array_delete(fit_part);
    }
    cpl_array_delete(par_all);

    // Compute and save the overall fitting profile on over-chunks
    cpl_array *fit   = cpl_array_new(size_over, CPL_TYPE_DOUBLE);
    for (int l = 0; l < (int)size_over; l++) {

        // If lines are more than one, compute overall profile; otherwise just
        // copy the individual profile
        if ((int)par_num > 4) { 
            cpl_array_set_double(fit, l, exp(-1.0 * 
                                 cpl_array_get_double(tot, l, NULL)));
        } 
        else {
            cpl_array_set_double(fit, l, 
                                 cpl_array_get_double(tot, l, NULL));
        } 
    }
    cpl_array_delete(tot);

    // Convolve the overall profile with the instrument profile on chunks
    cpl_array *fit_conv = espda_spec_conv(((struct gsl_multifit_data *)data)
        ->prof_chunk, wavel_over, fit);
    cpl_array_delete(wavel_over);
    cpl_array_delete(fit); 

    //#######
    cpl_table_copy_data_double(spec_chunk, "FIT", 
                               cpl_array_get_data_double(fit_conv));
    //spec_chunk = cpl_array_get_data_double(fit_conv);
    //#######
    /*
    for (int i = 0; i < (int)par_num / 4; i++) { 
        
        cpl_table_set_double(lfit_chunk, "PEAK", i, );
    }
    */

    // Create a GSL vector with the model
    //#######
    for (int i = 0; i < (int)size_chunk; i++) {
      //for (int i = 0; i < (int)cpl_array_get_size(fit_conv); i++) {      
    //#######

      double fiti     = cpl_array_get_double(fit_conv, i, NULL);

      //#######
        double normi    = cpl_table_get_double(spec_chunk, "FLUX",    i, NULL) /
                          cpl_table_get_double(spec_chunk, "CONT",    i, NULL);
        double normerri = cpl_table_get_double(spec_chunk, "FLUXERR", i, NULL) /
                          cpl_table_get_double(spec_chunk, "CONT",    i, NULL);
	//double normi     = cpl_array_get_double(spec_norm, i, NULL);
	//double normerri  = cpl_array_get_double(spec_normerr, i, NULL);
    //#######
        gsl_vector_set(model, i, (fiti - normi) / normerri);
    }
    cpl_array_delete(fit_conv);

    return GSL_SUCCESS;
}

/*----------------------------------------------------------------------------*/
/**

 */
/*----------------------------------------------------------------------------*/

double espda_model_voigt(double wavel_i, 
			 double ion_wavel,
			 double ion_oscs,
			 double ion_damp,
			 double linez,
			 double colden,
			 double therb,
			 double turbb,
			 cpl_table *voigt_func) {

    /*
    struct model_voigt_params *params = (struct model_voigt_params *)p;
    double ion_wavel = (params->ion_wavel) * 1.0e-9;  // nm to m
    double ion_oscs = (params->ion_oscs); 
    double ion_damp = (params->ion_damp);
    double linez = (params->linez);
    double colden = (params->colden) + 4.0;           // 1/cm2 to 1/m2
    double therb = (params->therb) * 1.0e3;           // km/s to m/s
    double turbb = (params->turbb) * 1.0e3;           // km/s to m/s
    cpl_table *voigt_func = (params->voigt_func);
    */          
    ion_wavel = ion_wavel * 1.0e-9;   // nm to m
    colden    = colden + 4.0;         // 1/cm2 to 1/m2
    therb     = therb * 1.0e3;        // km/s to m/s
    turbb     = turbb * 1.0e3;        // km/s to m/s
    wavel_i = wavel_i * 1.0e-9;       // nm to m
                   
    double b      = sqrt(therb * therb + turbb * turbb);    // Broadening par.
    double a      = 0.25 * ion_damp * ion_wavel *           // Argument a
                    CPL_MATH_1_PI / b; 
    double optd_0 = ion_oscs * pow(10, colden) * ERADIUS *  // Normalisation
                    CPL_MATH_SQRT2PI * CPL_MATH_SQRT1_2  * 
                    LIGHTSPEED1000 * ion_wavel / (b + 0  *
                    sqrt(2 * log(2)));
    double u = LIGHTSPEED1000 * (ion_wavel -            // Argument u
               wavel_i / (linez + 1)) / (b * ion_wavel);
    double voigt_value;    

    // Option 1: compute the Voigt function
    if (voigt_func == NULL) {
        int space = 1000;   

        // Compute integral
        struct espda_voigt_integ_params  params = {a, u};
        gsl_function                     integrand;
        gsl_integration_workspace       *w      = 
            gsl_integration_workspace_alloc(space);
        double                           integ_res, integ_err;
        integrand.function = &espda_voigt_integ;
        integrand.params   = &params;
        gsl_integration_qags(&integrand, 0, u, 1e-30, 1e-10, space, w, 
                             &integ_res, &integ_err);
        gsl_integration_workspace_free(w);

        if (u * u < 2500) {
        double erfcx;

            // Compute or approximate the term outside the integral
        	if (a < 26.6) {
        	    erfcx = exp(a * a) * erfc(a);
        	}
        	else {
        	    erfcx =  0.5 * CPL_MATH_2_SQRTPI * 1 / a *
        	             (1 - 0.5  / (a * a)    + 0.75       / pow(a, 3) - 
        	              1.875    / pow(a, 6)  + 6.5625     / pow(a, 8) - 
        	              29.53125 / pow(a, 10) + 162.421875 / pow(a, 12));
        	}
    
        	// Compute the function
        	voigt_value = erfcx * exp(-1 * u * u) * cos(2 * a * u) + 
        	              integ_res * CPL_MATH_2_SQRTPI;
        }
        else {
            voigt_value = a * CPL_MATH_2_SQRTPI / (2 * u * u);
        }
    }
    
    // Option 2: interpolate from the static table
    else {

        // Define the location of interpolation and the points to be used
        int    j_max = cpl_table_get_nrow(voigt_func);
        int    k_max = cpl_array_get_size(
                       cpl_table_get_array(voigt_func, "VOIGT_FUNC", 1));
        double j_0   = (cbrt(0.5 * 1e-4 * u) + 0.5) * j_max;
        double k_0   = ((log10(a) + 5) / 3) * k_max;
        int    j_a   = fmin(fmax((int)floor(j_0), 0), j_max - 2);
        int    j_b   = fmin(fmax((int)ceil (j_0), 1), j_max - 1);
        int    k_a   = fmin(fmax((int)floor(k_0), 0), k_max - 2);
        int    k_b   = fmin(fmax((int)ceil (k_0), 1), k_max - 1);
        double v_aa  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                       "VOIGT_FUNC", j_a), k_a, NULL);
        double v_ba  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                       "VOIGT_FUNC", j_b), k_a, NULL);
        double v_ab  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                       "VOIGT_FUNC", j_a), k_b, NULL);
        double v_bb  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                       "VOIGT_FUNC", j_b), k_b, NULL);

        // Interpolate
        voigt_value = espda_interp_bilin(j_a, j_b, k_a, k_b, v_aa, v_ba, 
            v_ab, v_bb, j_0, k_0);
    }
    return voigt_value * optd_0;
}

/*----------------------------------------------------------------------------*/
/**

 */
/*----------------------------------------------------------------------------*/

/*
double espda_model_voigtf(double x, void *p) {
  
    struct model_voigt_params *params = (struct model_voigt_params *)p;
    const char *px =  ((struct model_voigtf_params *)params)->px;
    double wavel_i = ((struct model_voigtf_params *)params)->wavel_i;
    double ion_wavel = ((struct model_voigtf_params *)params)->ion_wavel;
    double ion_oscs = ((struct model_voigtf_params *)params)->ion_oscs; 
    double ion_damp = ((struct model_voigtf_params *)params)->ion_damp;
    double linez = ((struct model_voigtf_params *)params)->linez;
    double colden = ((struct model_voigtf_params *)params)->colden;
    double therb = ((struct model_voigtf_params *)params)->therb;
    double turbb = ((struct model_voigtf_params *)params)->turbb;
    cpl_table *voigt_func = ((struct model_voigtf_params *)params)->voigt_func;
    double voigt_value;
    if (!strcmp(px, "linez")) {
        voigt_value = espda_model_voigt(wavel_i, ion_wavel, ion_oscs, ion_damp,
					x, colden, therb, turbb, voigt_func);
    }
    else if (!strcmp(px, "colden")){
        voigt_value = espda_model_voigt(wavel_i, ion_wavel, ion_oscs, ion_damp,
					linez, x, therb, turbb, voigt_func);
    }
    else if (!strcmp(px, "therb")){
        voigt_value = espda_model_voigt(wavel_i, ion_wavel, ion_oscs, ion_damp,
					linez, colden, x, turbb, voigt_func);
    }
    else if (!strcmp(px, "turbb")){
        voigt_value = espda_model_voigt(wavel_i, ion_wavel, ion_oscs, ion_damp,
					linez, colden, therb, x, voigt_func);
    }
    else {
        voigt_value = INV_DOUBLE;
    }
    return voigt_value;
}
*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Model the Voigt function
  @param    wavel       Wavelength domain
  @param    ion_wavel   Wavelength of the ionic transition
  @param    ion_oscs    Oscillator strength of the ionic transition
  @param    ion_damp    Damping constant of the ionic transition
  @param    linez       Redshift of the line
  @param    colden      Column density of the line
  @param    therb       Thermal broadening of the line
  @param    turbb       Turbulence broadening of the line
  @param    voigt_func  Values of the Voigt function
  @param    voigt       Voigt profile
  @param    fit         Fitting profile
  @return   CPL_ERROR_NONE iff OK  

  This function computes the Voigt function for an absorption line across a 
  given wavelength domain. The arguments of the Voigt function are the 
  parameters of the line; other parameters are those of the associated ionic 
  transition. Two profiles are given in output, the Voigt profile (i.e. the 
  Voigt function computed at each point of the wavelength domain) and the 
  fitting profile (i.e. the negative exponential of the Voigt profile, which 
  fits the actual transmission profile of the line).

  The Voigt function is expressed using the integral representation of Zaghloul 
  2007, which is exactly computable except for very small values of line 
  broadening. In this case, an asymptotic expansion (accurate to 16 significant 
  digits) is used instead. If a table with the values of the Voigt function is
  provided, the Voigt function is not computed but is interpolated from the 
  table.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_model_voigt_old(cpl_array  *wavel,
                                 double      ion_wavel,
                                 double      ion_oscs,
                                 double      ion_damp,
                                 double      linez,
                                 double      colden,
                                 double      therb,
                                 double      turbb,
                                 cpl_table  *voigt_func,
                                 cpl_array **voigt,
                                 cpl_array **fit,
                                 double     *peak)  

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Adjust variables and define parameters
    ion_wavel = ion_wavel * 1.0e-9;                         // nm to m
    colden    = colden + 4.0;                  		    // 1/cm2 to 1/m2
    therb     = therb * 1.0e3;                 		    // km/s to m/s
    turbb     = turbb * 1.0e3;                 		    // km/s to m/s
    cpl_array_multiply_scalar(wavel, 1.0e-9);  		    // nm to m
    double b      = sqrt(therb * therb + turbb * turbb);    // Broadening par.
    double a      = 0.25 * ion_damp * ion_wavel *           // Argument a
                    CPL_MATH_1_PI / b; 
    double optd_0 = ion_oscs * pow(10, colden) * ERADIUS *  // Normalisation
                    CPL_MATH_SQRT2PI * CPL_MATH_SQRT1_2  * 
                    LIGHTSPEED1000 * ion_wavel / (b + 0  *
                    sqrt(2 * log(2)));
    //printf("%e %e %e %f %f\n", colden, therb, ion_wavel, optd_0, a);

    // Prepare integration
    cpl_size wavel_size = cpl_array_get_size(wavel);
    *fit                = cpl_array_new(wavel_size, CPL_TYPE_DOUBLE);
    *voigt              = cpl_array_new(wavel_size, CPL_TYPE_DOUBLE);
    int space = 1000;   

    // Loop over positions in wavelength domain
    for (int i = 0; i < wavel_size; i++) {
        double u = LIGHTSPEED1000 * (ion_wavel -            // Argument u
                   cpl_array_get_double(wavel, i, NULL) / 
                   (linez + 1)) / (b * ion_wavel);
        double voigt_value;

        // Option 1: compute the Voigt function
        if (voigt_func == NULL) {

            // Compute integral
            struct espda_voigt_integ_params  params = {a, u};
            gsl_function                     integrand;
            gsl_integration_workspace       *w      = 
                gsl_integration_workspace_alloc(space);
            double                           integ_res, integ_err;
            integrand.function = &espda_voigt_integ;
            integrand.params   = &params;
            gsl_integration_qags(&integrand, 0, u, 1e-30, 1e-10, space, w, 
                                 &integ_res, &integ_err);
            gsl_integration_workspace_free(w);

            if (u * u < 2500) {
            double erfcx;

                // Compute or approximate the term outside the integral
            	if (a < 26.6) {
            	    erfcx = exp(a * a) * erfc(a);
            	}
            	else {
            	    erfcx =  0.5 * CPL_MATH_2_SQRTPI * 1 / a *
            	             (1 - 0.5  / (a * a)    + 0.75       / pow(a, 3) - 
            	              1.875    / pow(a, 6)  + 6.5625     / pow(a, 8) - 
            	              29.53125 / pow(a, 10) + 162.421875 / pow(a, 12));
            	}
     
            	// Compute the function
            	voigt_value = erfcx * exp(-1 * u * u) * cos(2 * a * u) + 
            	              integ_res * CPL_MATH_2_SQRTPI;
            }
            else {
                voigt_value = a * CPL_MATH_2_SQRTPI / (2 * u * u);
            }
        }
       
        // Option 2: interpolate from the static table
        else {

            // Define the location of interpolation and the points to be used
            int    j_max = cpl_table_get_nrow(voigt_func);
            int    k_max = cpl_array_get_size(
                           cpl_table_get_array(voigt_func, "VOIGT_FUNC", 1));
            double j_0   = (cbrt(0.5 * 1e-4 * u) + 0.5) * j_max;
            double k_0   = ((log10(a) + 5) / 3) * k_max;
            int    j_a   = fmin(fmax((int)floor(j_0), 0), j_max - 2);
            int    j_b   = fmin(fmax((int)ceil (j_0), 1), j_max - 1);
            int    k_a   = fmin(fmax((int)floor(k_0), 0), k_max - 2);
            int    k_b   = fmin(fmax((int)ceil (k_0), 1), k_max - 1);
            double v_aa  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                           "VOIGT_FUNC", j_a), k_a, NULL);
            double v_ba  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                           "VOIGT_FUNC", j_b), k_a, NULL);
            double v_ab  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                           "VOIGT_FUNC", j_a), k_b, NULL);
            double v_bb  = cpl_array_get_double(cpl_table_get_array(voigt_func,
                           "VOIGT_FUNC", j_b), k_b, NULL);

            // Interpolate
            voigt_value = espda_interp_bilin(j_a, j_b, k_a, k_b, v_aa, v_ba, 
                v_ab, v_bb, j_0, k_0);
        }
        cpl_array_set_double(*voigt, i,      optd_0 * voigt_value);
        cpl_array_set_double(*fit,   i, exp(-optd_0 * voigt_value));
    }
    cpl_array_multiply_scalar(wavel, 1e9);  // m to nm
    
    //*peak = cpl_array_get_max(*voigt);
    *peak = cpl_array_get_min(*fit);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to compute the Voigt function.");
    }
    return CPL_ERROR_NONE;
}

/**@}*/
