/* $Id: espda_compu_rhk.c 193686 2016-11-28 16:27:52Z asegovia $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: asegovia $
 * $Date: 2016-11-28 17:27:52 +0100 (Tue, 22 Nov 2016) $
 * $Revision: 193686 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_rhk.h"
#include "espda_ccf.h"
#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_param.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_compu_rhk_create(cpl_plugin *);
static int espda_compu_rhk_exec(cpl_plugin *);
static int espda_compu_rhk_destroy(cpl_plugin *);
static int espda_compu_rhk(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_compu_rhk_description[] =
"This recipe compute the activity indexes S and R'HK from a stellar spectrum.\n"
"\n"
"INPUT FRAMES        FORMAT  # EXPECTED                                  "
"          TAGS\n"
"equalized spectrum  SPEC    0 to 1                                      "
"          EQ_SPEC\n"
"rebinned  spectrum  SPEC    0 to 1;                                     "
"          REBIN_SPEC\n"
"order     spectrum  SPEC    0 to 1; 1 if EQ_SPEC and REBIN_SPEC are not "
"provided  ORD_SPEC\n"
"\n"
"OUTPUT FRAMES       FORMAT  # PRODUCED                                  "
"          TAGS\n"
"equalized spectrum  SPEC    same as EQ_SPEC                             "
"          EQ_SPEC_MASK\n"
"rebinned  spectrum  SPEC    same as REBIN_SPEC                          "
"          REBIN_SPEC_MASK\n"
"order     spectrum  SPEC    same as ORD_SPEC                            "
"          ORD_SPEC_MASK\n"
"mask                MASK    1                                           "
"          MASK"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_compu_rhk", "Compute "
                        "the activity indexes S and RHK from a stellar spec",
                        espda_compu_rhk_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_compu_rhk_create, espda_compu_rhk_exec,
                        espda_compu_rhk_destroy)) {
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    if (cpl_pluginlist_append(list, plugin)) {
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);
        return 1;
    }

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_rhk_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;
    cpl_parameter *param;

    /* Do not create the recipe if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Parameter list allocation failed");
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);
    }

    /* Fill the parameters list */

/*
 * Default values for all parameters are set to INV_DOUBLE = 9999.0, because no
 * other default has been determined yet.
 */

    param = cpl_parameter_new_value("espda.espda_compu_rhk.bv",
        CPL_TYPE_DOUBLE, "B-V color of the star",
        "espda.espda_compu_rhk", 0.82);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "bv");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_rhk.rv",
        CPL_TYPE_DOUBLE, "Star's RV",
        "espda.espda_compu_rhk", 9999.);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "rv");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_rhk.ca_alpha",
        CPL_TYPE_DOUBLE, "Calcium index alpha",
        "espda.espda_compu_rhk", 1.044);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ca_alpha");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_rhk.ca_beta",
        CPL_TYPE_DOUBLE, "Calcium index beta",
        "espda.espda_compu_rhk", 0.031);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "ca_beta");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_compu_rhk.adp_columns",
				    CPL_TYPE_STRING,
				    "Columns to upload from input in the S1D ADP format (disregarded "
            "if input has a different format)",
				    "espda.espda_compu_rhk",
				    "WAVE_AIR,FLUX,ERR,QUAL");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "adp-columns");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_rhk_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();

    /* Return immediately if an error code is already set */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());
        return (int)cpl_error_get_code();
    }

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    /* Verify parameter and frame lists */
    if (recipe->parameters == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }
    if (recipe->frames == NULL) {
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Invoke the recipe */
    recipe_status = espda_compu_rhk(recipe->frames, recipe->parameters);

    /* Ensure DFS-compliance of the products */
    if (cpl_dfs_update_product_header(recipe->frames)) {
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();
    }

    if (!cpl_errorstate_is_equal(initial_errorstate)) {
        /* Dump the error history since recipe execution start.
           At this point the recipe cannot recover from the error */
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);
    }

    return recipe_status;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_rhk_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;

    if (plugin == NULL) {
        cpl_msg_error(cpl_func, "Null plugin");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);
    }

    /* Verify plugin type */
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {
        cpl_msg_error(cpl_func, "Plugin is not a recipe");
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);
    }

    /* Get the recipe */
    recipe = (cpl_recipe *)plugin;

    cpl_parameterlist_delete(recipe->parameters);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_rhk(cpl_frameset            *frameset,
                           const cpl_parameterlist *parlist) {

/*
 * Input is checked and loaded.
 */
	/* check for invalid input */
	if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
		return CPL_ERROR_BAD_FILE_FORMAT;
	}
    espda_msg("Starting rhk computation");

    // Load and check frames
    espda_dfs_ensure(espda_dfs_groups(frameset));

    cpl_error_code my_error = CPL_ERROR_NONE;

    /* Extracting the S1D frame from the SOF */
    cpl_frameset *S1D_frameset = cpl_frameset_new();
    my_error = espda_frame_extract_by_tag(frameset, "S1D", S1D_frameset);
    const char *s1d_name = cpl_frame_get_filename(cpl_frameset_get_position(S1D_frameset, 0));
    espda_msg("s1d_name = %s", s1d_name);
    cpl_table *temp = cpl_table_load(s1d_name,1,0);
    cpl_table *S1D_table = NULL;

    char *adp_columns = (char *)malloc(sizeof(char) * LEN_STRING);;
    /*
    double bv, rv, ca_alpha, ca_beta;
    espda_param_load(parlist, "espda.espda_compu_rhk.bv", &bv, -1.0, 1.0);
    espda_param_load(parlist, "espda.espda_compu_rhk.rv", &rv, -100000.0, 100000.0);
    espda_param_load(parlist, "espda.espda_compu_rhk.ca_alpha", &ca_alpha, -2.0, 2.0);
    espda_param_load(parlist, "espda.espda_compu_rhk.ca_beta", &ca_beta, -2.0, 2.0);
    */
    espda_param_load(parlist, "espda.espda_compu_rhk.adp_columns", &adp_columns, "", "");

    // Choosing the columns for ADP products
    cpl_array *cols = espda_param_extract(adp_columns, ",", 0, 1, CPL_TYPE_STRING);
    if (cpl_table_get_nrow(temp)==1) {
      const char *wavel_col = cpl_array_get_string(cols, 0);
      const char *flux_col = cpl_array_get_string(cols, 1);
      const char *fluxerr_col = cpl_array_get_string(cols, 2);
      const char *qualdata_col = cpl_array_get_string(cols, 3);
      const cpl_array *wavel_air = NULL, *wavel = NULL, *flux = NULL,
          *fluxerr = NULL, *qualdata = NULL;
      if (cpl_table_has_column(temp, wavel_col)) {
          wavel_air = cpl_table_get_array(temp, wavel_col, 0);
      }
      else {
          cpl_msg_error(cpl_func, "Column %s does not exist.", wavel_col);
      }
      wavel = cpl_table_get_array(temp, "WAVE", 0);
      //cpl_array_divide_scalar(wavel, 10.0);
      if (cpl_table_has_column(temp, flux_col)) {
          flux = cpl_table_get_array(temp, flux_col, 0);
      }
      else {
          cpl_msg_error(cpl_func, "Column %s does not exist.", flux_col);
      }
      if (cpl_table_has_column(temp, fluxerr_col)) {
          fluxerr = cpl_table_get_array(temp, fluxerr_col, 0);
      }
      else {
          cpl_msg_error(cpl_func, "Column %s does not exist.", fluxerr_col);
      }
      if (cpl_table_has_column(temp, qualdata_col)) {
          qualdata = cpl_table_get_array(temp, qualdata_col, 0);
      }
      else {
          cpl_msg_error(cpl_func, "Column %s does not exist.", qualdata_col);
      }

      /*
      const cpl_array *wavel = cpl_table_get_array(temp, "WAVE", 0);
      const cpl_array *flux = cpl_table_get_array(temp, "FLUX", 0);
      const cpl_array *fluxerr = cpl_table_get_array(temp, "ERR", 0);
      const cpl_array *qualdata = cpl_table_get_array(temp, "QUAL", 0);
      const cpl_array *wavel_air = cpl_table_get_array(temp, "WAVE_AIR", 0);
      */
      cpl_size size_in = cpl_array_get_size(wavel);
      S1D_table = cpl_table_new(size_in);
      cpl_table_new_column(S1D_table, "wavelength", CPL_TYPE_DOUBLE);
      cpl_table_new_column(S1D_table, "wavelength_air", CPL_TYPE_DOUBLE);
      cpl_table_new_column(S1D_table, "flux", CPL_TYPE_DOUBLE);
      cpl_table_new_column(S1D_table, "error", CPL_TYPE_DOUBLE);
      cpl_table_new_column(S1D_table, "quality", CPL_TYPE_INT);
      cpl_table_copy_data_double(S1D_table, "wavelength",
        cpl_array_get_data_double_const(wavel));
      cpl_table_copy_data_double(S1D_table, "wavelength_air",
        cpl_array_get_data_double_const(wavel_air));
      cpl_table_copy_data_double(S1D_table, "flux",
        cpl_array_get_data_double_const(flux));
      cpl_table_copy_data_double(S1D_table, "error",
        cpl_array_get_data_double_const(fluxerr));
      cpl_table_copy_data_int(S1D_table, "quality",
        cpl_array_get_data_int_const(qualdata));
    }
    else {
        S1D_table = cpl_table_duplicate(temp);
    }
    cpl_table_delete(temp);

    /* Getting keywords from the frame */
    cpl_propertylist *keywords = NULL;
    keywords = cpl_propertylist_load(cpl_frame_get_filename(
                                cpl_frameset_get_position(S1D_frameset, 0)), 0);
    const char *INSTRUME = cpl_propertylist_get_string(keywords, "INSTRUME");

    double pbflux_k,pbflux_k_err;
    double pbflux_h,pbflux_h_err;
    double pbflux_v,pbflux_v_err;
    double pbflux_r,pbflux_r_err;
    double s_raw, s_err, s_mw;
    double rhk, rhk_err;
    double prot,prot_err,age,age_err;

    int S1D_length = cpl_table_get_nrow(S1D_table);
    int sci_table_cols = cpl_table_get_ncol(S1D_table);

    cpl_image *flux, *error, *quality;

    espda_param_double("bv",  bv,
        cpl_parameterlist_get_first_const(parlist), -1.0, 1.0);
    espda_param_double("rv",  rv,
        cpl_parameterlist_get_next_const(parlist), -100000.0, 100000.0);
    espda_param_double("ca_alpha",  ca_alpha,
        cpl_parameterlist_get_next_const(parlist), -2.0, 2.0);
    espda_param_double("ca_beta",  ca_beta,
        cpl_parameterlist_get_next_const(parlist), -2.0, 2.0);


    /* Overload parameters */
    const cpl_parameter * par_bv =
         cpl_parameterlist_find_const(parlist,"espda.espda_compu_rhk.bv");
    int bv_set = cpl_parameter_get_default_flag(par_bv);

    const cpl_parameter * par_rv =
         cpl_parameterlist_find_const(parlist,"espda.espda_compu_rhk.rv");
    int rv_set = cpl_parameter_get_default_flag(par_rv);

    const cpl_parameter * par_ca_alpha =
         cpl_parameterlist_find_const(parlist,"espda.espda_compu_rhk.ca_alpha");
    int ca_alpha_set = cpl_parameter_get_default_flag(par_ca_alpha);

    const cpl_parameter * par_ca_beta =
         cpl_parameterlist_find_const(parlist,"espda.espda_compu_rhk.ca_beta");
    int ca_beta_set = cpl_parameter_get_default_flag(par_ca_beta);

    if(bv_set==1) {
            espda_param_double("bv",bv,
            cpl_parameterlist_find_const(parlist, "espda.espda_compu_rhk.bv"),
                               -1.0, 1.0);
    }

    if(rv_set==1) {
            espda_param_double("rv",rv,
            cpl_parameterlist_find_const(parlist, "espda.espda_compu_rhk.rv"),
                               -100000.0, 100000.0);
    }

    if(ca_alpha_set==1) {
            espda_param_double("ca_alpha",ca_alpha,
            cpl_parameterlist_find_const(parlist, "espda.espda_compu_rhk.ca_alpha"),
                               -2.0, 2.0);
    }

    if(ca_beta_set==1) {
            espda_param_double("ca_beta",ca_beta,
            cpl_parameterlist_find_const(parlist, "espda.espda_compu_rhk.ca_beta"),
                               -2.0, 2.0);
    }



    /* Picking the right RADVEL */

    double RADVEL;

    if (rv != 9999.0) {
        RADVEL = rv;
    } else {
        RADVEL = cpl_propertylist_get_double(keywords,"ESO OCS OBJ RV");
    }

    espda_msg(ANSI_COLOR_GREEN"bv = %f"ANSI_COLOR_RESET, bv);
    espda_msg(ANSI_COLOR_GREEN"rv = %f"ANSI_COLOR_RESET, RADVEL);
    espda_msg(ANSI_COLOR_GREEN"ca_alpha = %f"ANSI_COLOR_RESET, ca_alpha);
    espda_msg(ANSI_COLOR_GREEN"ca_beta = %f"ANSI_COLOR_RESET, ca_beta);

    /* CLO: shifting wavelength array to stellar rest frame */
	double *wave_shifted;
	wave_shifted = cpl_table_get_data_double(S1D_table, "wavelength_air");
    for (int j=0; j < S1D_length; j++) {
        wave_shifted[j] = wave_shifted[j]/(1.+RADVEL/SPEED_OF_LIGHT);
    }

    /* Compute the flux density in a given passband */
    espda_msg("Executing espda_measure_flux_HK band K...");
    my_error = espda_measure_flux_HK(wave_shifted,
                                     cpl_table_get_data_double(S1D_table,"flux"),
                                     cpl_table_get_data_double(S1D_table,"error"),
                                     START_WAVELENGTH_K,
                                     END_WAVELENGTH_K,
                                     CENTRAL_WAVELENGTH_K,
                                     WIDTH_K,
                                     cpl_table_get_data_int(S1D_table,"quality"),
                                     S1D_length,
                                     &pbflux_k,
                                     &pbflux_k_err);

    espda_msg("pbflux_k     = %f", pbflux_k);
    espda_msg("pbflux_k_err = %f", pbflux_k_err);

    /* Compute the flux density in a given passband */
    espda_msg("Executing espda_measure_flux_HK band H...");
    my_error = espda_measure_flux_HK(wave_shifted,
                                     cpl_table_get_data_double(S1D_table,"flux"),
                                     cpl_table_get_data_double(S1D_table,"error"),
                                     START_WAVELENGTH_H,
                                     END_WAVELENGTH_H,
                                     CENTRAL_WAVELENGTH_H,
                                     WIDTH_H,
                                     cpl_table_get_data_int(S1D_table,"quality"),
                                     S1D_length,
                                     &pbflux_h,
                                     &pbflux_h_err);

    espda_msg("pbflux_h     = %f", pbflux_h);
    espda_msg("pbflux_h_err = %f", pbflux_h_err);

    /* Compute the flux density in a given passband */
    espda_msg("Executing espda_measure_flux band V...");
    my_error = espda_measure_flux(wave_shifted,
                                  cpl_table_get_data_double(S1D_table,"flux"),
                                  cpl_table_get_data_double(S1D_table,"error"),
                                  START_WAVELENGTH_V,
                                  END_WAVELENGTH_V,
                                  cpl_table_get_data_int(S1D_table,"quality"),
                                  S1D_length,
                                  &pbflux_v,
                                  &pbflux_v_err);

    espda_msg("pbflux_v     = %f", pbflux_v);
    espda_msg("pbflux_v_err = %f", pbflux_v_err);

    /* Compute the flux density in a given passband */
    espda_msg("Executing espda_measure_flux band R...");
    my_error = espda_measure_flux(wave_shifted,
                                  cpl_table_get_data_double(S1D_table,"flux"),
                                  cpl_table_get_data_double(S1D_table,"error"),
                                  START_WAVELENGTH_R,
                                  END_WAVELENGTH_R,
                                  cpl_table_get_data_int(S1D_table,"quality"),
                                  S1D_length,
                                  &pbflux_r,
                                  &pbflux_r_err);

    espda_msg("pbflux_r      = %f", pbflux_r);
    espda_msg("pbflux_r_err  = %f", pbflux_r_err);

    s_raw = (pbflux_h + pbflux_k) / (pbflux_v + pbflux_r);

    //s_err = sqrt(pow((pbflux_h_err+pbflux_k_err)/(pbflux_v+pbflux_r),2.0) +
	      //pow(pbflux_h+pbflux_k,2.0)*(pbflux_v_err+pbflux_r_err) /
	      //pow(pbflux_v+pbflux_r,4.0));

    s_err = (pbflux_h_err*pbflux_h_err+pbflux_k_err*pbflux_k_err)/(pbflux_v + pbflux_r)/(pbflux_v + pbflux_r) +
	    (pbflux_h + pbflux_k)*(pbflux_h + pbflux_k)*(pbflux_v_err*pbflux_v_err+pbflux_r_err*pbflux_r_err) /
	    pow((pbflux_v + pbflux_r),4.0);

    s_err = sqrt(s_err);

    espda_msg("s_raw = %f", s_raw);
    espda_msg("s_err = %f", s_err);

    s_mw = ca_alpha*s_raw + ca_beta;
    s_err = ca_alpha*s_err;

    espda_msg("s_mw  = %f", s_mw);
    espda_msg("s_err = %f", s_err);

    /* Compute the RHK activity index */
    espda_msg("Executing espda_compu_rhk_index...");
    my_error = espda_compu_rhk_index(s_mw,
                                     s_err,
                                     bv,
                                     &rhk,
                                     &rhk_err);

    espda_msg("rhk     = %f", rhk);
    espda_msg("rhk_err = %f", rhk_err);

    espda_msg("Executing espda_compu_prot_age...");
    my_error = espda_compu_prot_age(rhk,
                                    rhk_err,
                                    bv,
                                    &prot,
                                    &prot_err,
                                    &age,
                                    &age_err);

    espda_msg("prot     = %f", prot);
    espda_msg("prot_err = %f", prot_err);
    espda_msg("age      = %f", age);
    espda_msg("age_err  = %f", age_err);

/*
 * Output is exported. QC parameters are saved.
 */

    espda_msg("Saving products with QCs ...");

    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII BV", bv);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII SRAW", s_raw);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII SMW", s_mw);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII SMW ERR", s_err);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII RHK", rhk);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII RHK ERR", rhk_err);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII PROT", prot);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII PROT ERR", prot_err);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII AGE", age);
    cpl_propertylist_append_double(keywords, "ESO QC RHK CAII AGE ERR", age_err);

    cpl_propertylist_update_string(keywords, CPL_DFS_PRO_TYPE, "ANALYZED");
    cpl_propertylist_update_string(keywords, CPL_DFS_PRO_CATG, "RHK_FLUX");

    char s1d_analyzed_name[64];
    sprintf(s1d_analyzed_name,"%s_S1D_RHK.fits", INSTRUME);
    my_error = cpl_dfs_save_table(frameset, NULL, parlist, frameset,
                                  NULL, S1D_table, NULL,
                                  "espda_compu_rhk", keywords, NULL, PACKAGE "/"
                                  PACKAGE_VERSION, s1d_analyzed_name);

    espda_ensure(my_error != CPL_ERROR_NONE, my_error,
                 "Saving table %s failed: %s", s1d_analyzed_name,
                 cpl_error_get_message_default(my_error));

    /* Cleaning memory */
    espda_msg("Cleaning memory");
    cpl_frameset_delete(S1D_frameset);
    cpl_propertylist_delete(keywords);
    cpl_table_delete(S1D_table);

    espda_msg("Rhk computation finished correctly");

    return (int)cpl_error_get_code();
}

/**@}*/
