/* $Id: espda_fit_qsocont.c 167090 2014-10-31 16:35:06Z gcupani $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: gcupani $
 * $Date: 2014-10-31 17:35:06 +0100 (Fri, 31 Oct 2014) $
 * $Revision: 167090 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif


#include <time.h>

#ifdef _OPENMP
#  include <omp.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "espda_check.h"
#include "espda_dfs.h"
#include "espda_fit.h"
#include "espda_frame.h"
#include "espda_interp.h"
#include "espda_line.h"
#include "espda_model.h"
#include "espda_param.h"      
#include "espda_split.h"
#include "espda_spec.h"
#include "espda_struct.h"
#include "espda_voigt.h"
#include "espda_wrap.h"
//#include "sgfilter.h"

#include <cpl.h>
#include <gsl/gsl_bspline.h>
#include <gsl/gsl_multifit.h>

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_fit_qsocont_create(cpl_plugin *);
static int espda_fit_qsocont_exec(cpl_plugin *);
static int espda_fit_qsocont_destroy(cpl_plugin *);
static int espda_fit_qsocont(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_fit_qsocont_description[] =
"This recipe determines the emission continuum in a QSO spectrum. The "
"continuum is iteratively determined by removing the absorption lines redwards "
"and bluewards from the Lyman-alpha emission line.\n"
"\n"
"INPUT FRAMES                          FORMAT  # EXPECTED  TAGS        "
"NOTES\n" 
"Full spectrum                         SPEC    0 or 1      FSPEC\n"
"Rebinned  spectrum                    SPEC    1           RSPEC\n"
"List of absorption lines              LINE    0 or 1      FLINE_PRE\n"
"List of ionic transitions             ION     1           QSO_ION\n"
"Table with Voigt function             VOIGT   0 or 1      VOIGT_FUNC  "
"recommended (speeds up computation)\n"
"Rebinned instrument profile           PROF    0 or 1      RPROF\n"
"\n"
"OUTPUT FRAMES                         FORMAT  # PRODUCED  TAGS        "
"NOTES\n" 
"Full spectrum with continuum          SPEC    0 or 1      FSPEC_CONT  "
"Produced only if FSPEC is provided\n"
"Full spectrum with lines removed      SPEC    0 or 1      FSPEC_REM   "
"Produced only if FSPEC is provided\n"
"Rebinned spectrum with continuum      SPEC    1           RSPEC_CONT\n"
"Rebinned spectrum with lines removed  SPEC    1           RSPEC_REM\n"
"List of removed absorption lines      LINE    0 or 1      FLINE_REM   "
"produced only if FLINE_PRE is provided\n"
"Table with spline points              SPLINE  1           SPLINE_CONT "
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_fit_qsocont", "Fit the "
                        "emission continuum in a QSO spectrum", 
                        espda_fit_qsocont_description, "Guido Cupani",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_fit_qsocont_create, espda_fit_qsocont_exec,
                        espda_fit_qsocont_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_qsocont_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;                                               
    cpl_parameter *param;
                                                                     
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */
    param = cpl_parameter_new_value("espda.espda_fit_qsocont.hwidth", 
				    CPL_TYPE_DOUBLE,
				    "Halfwidth for line grouping [nm]: "
				    "[2.5e-2, 1.0e1]",
				    "espda.espda_fit_qsocont",
				    0.05);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "hwidth");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.par_range",
				    CPL_TYPE_STRING,
				    "Range of line parameters", 
				    "espda.espda_fit_qsocont",
				    "1.0e-5,10.0,17.0,2.0,50.0,0.0,0.0");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "par-range");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.peak_thres",
				    CPL_TYPE_DOUBLE,
				    "Relative threshold to cut line peaks "
				    "during normalization: [1.0e-2, 1.0]",
				    "espda.espda_fit_qsocont",
				    0.8);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "peak-thres");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.vel_sampl_blue",
				    CPL_TYPE_DOUBLE,
				    "Velocity sampling to fit a spline to the "
				    "blue part [km/s]: [1.0e2, 1.0e4]",
				    "espda.espda_fit_qsocont", 
				    2000.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "vel-sampl-blue");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.vel_sampl_red",
				    CPL_TYPE_DOUBLE,
				    "Velocity sampling to fit a spline to the "
				    "red part [km/s]: [1.0e2, 1.0e4]",
				    "espda.espda_fit_qsocont", 
				    1000.0);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "vel-sampl-red");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    /*
    param = cpl_parameter_new_value("espda.espda_fit_qsocont.sg_width_pieces",
                CPL_TYPE_INT, "Width of the Savitzky-Golay filter used to "
		"smooth the continuum in pieces [px]: [1e2, 1e4]", 
		"espda.espda_fit_qsocont", 1000);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "sg-width-pieces");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.sg_width_final",
                CPL_TYPE_INT, "Width of the Savitzky-Golay filter used to "
		"smooth the final continuum [px]: [1e2, 1e4]", 
		"espda.espda_fit_qsocont", 500);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "sg-width-final");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
    */
    param = cpl_parameter_new_value("espda.espda_fit_qsocont.model_resol",
				    CPL_TYPE_DOUBLE,
				    "Resolution of the instrument model: "
				    "[1.0e3, 1.0e6]",
				    "espda.espda_fit_qsocont",
				    6.0e4);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "model-resol");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.iter_max_temp",
				    CPL_TYPE_INT,
				    "Maximum number of iterations in the blue "
				    "part: [-1, 1e3]",
				    "espda.espda_fit_qsocont",
				    10);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "iter-max");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    param = cpl_parameter_new_value("espda.espda_fit_qsocont.spline_knots",
				    CPL_TYPE_STRING,
				    "Manually-entered spline knots",
				    "espda.espda_fit_qsocont",
				    "");
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "spline-knots");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_qsocont_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe;                                                   
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();              
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = espda_fit_qsocont(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();                         
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_qsocont_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_fit_qsocont(cpl_frameset *frameset, const cpl_parameterlist 
                             *parlist) {
  //cpl_errorstate  prestate = cpl_errorstate_get();
    extern int v_level;
    extern char *grp_mode;

/*
 * Frames are loaded and checkeds.
 */
    /* check for invalid input */
    if(espda_files_dont_exist(frameset) != CPL_ERROR_NONE) {
    	return CPL_ERROR_BAD_FILE_FORMAT;
    }

    cpl_msg_info(cpl_func, "*** Load frames ***");
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());
    //cpl_size fspec_num = cpl_frameset_count_tags(frameset, FSPEC);
    cpl_recipeconfig *cfg = cpl_recipeconfig_new();
    char *cfg_tag = FSPEC;
    cpl_recipeconfig_set_tag  (cfg, cfg_tag, 0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, RSPEC,      1, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, FLINE_PRE,  0, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, QSO_ION,    1, 1);
    cpl_recipeconfig_set_input(cfg, cfg_tag, VOIGT_FUNC, 0, 1);
    //cpl_recipeconfig_set_input(cfg, cfg_tag, FPROF, 0, fspec_num);
    cpl_recipeconfig_set_input(cfg, cfg_tag, RPROF, 0, 1);
    espda_wrap_check(frameset, cfg, cfg_tag);
    cpl_table *fspec      = espda_frame_load(frameset, cfg, cfg_tag,	-1);
    cpl_table *rspec      = espda_frame_load(frameset, cfg, cfg_tag,	 0);
    cpl_table *fline_pre  = espda_frame_load(frameset, cfg, cfg_tag,	 1);
    cpl_table *ion        = espda_frame_load(frameset, cfg, cfg_tag,	 2);
    cpl_table *voigt_func = espda_frame_load(frameset, cfg, cfg_tag,	 3);
    //cpl_table *fprof      = espda_frame_load(frameset, cfg, cfg_tag,	 4);
    cpl_table *rprof      = espda_frame_load(frameset, cfg, cfg_tag,	 4);
    cpl_recipeconfig_delete(cfg);
    espda_util_safe(fspec, espda_spec_fit(fspec));
    espda_spec_fit(rspec);
    
/*
 * Parameters are loaded and checked.
 */

    cpl_msg_info(cpl_func, "*** Load parameters ***");
    espda_param_double("hwidth",         hwidth,    
        cpl_parameterlist_get_first_const(parlist), 2.5e-2, 1.0e1);
    espda_param_string("par-range",      par_range,    
        cpl_parameterlist_get_next_const(parlist));
    espda_param_double("peak-thres",     peak_thres,  
        cpl_parameterlist_get_next_const(parlist),  1.0e-2, 9.9e-1); 
    espda_param_double("vel-sampl-blue", vel_sampl_blue,   
        cpl_parameterlist_get_next_const(parlist),  1.0e2,  1.0e4);
    espda_param_double("vel-sampl-red",  vel_sampl_red,   
        cpl_parameterlist_get_next_const(parlist),  1.0e2,  1.0e4);
    /*
    espda_param_int   ("sg-width-pieces", sg_width_pieces,
        cpl_parameterlist_get_next_const(parlist),  1e2,    1e4);
    printf("%i\n", sg_width_pieces);
    espda_param_int   ("sg-width-final", sg_width_final,
        cpl_parameterlist_get_next_const(parlist),  1e2,    1e4);
    printf("%i\n", sg_width_final);
    */
    espda_param_double("model-resol",    model_resol,   
        cpl_parameterlist_get_next_const(parlist),  1.0e3,  1.0e6);
    espda_param_int   ("iter-max",       iter_max_temp,   
        cpl_parameterlist_get_next_const(parlist),  -1,     1e3);
    espda_param_string("spline-knots",   spline_file,    
        cpl_parameterlist_get_next_const(parlist));
    
/*
 * The spline knots are parsed, if provided, and converted into a table
 */

    char *spline_knots = (char *)malloc(sizeof(char) * LEN_STRING); 
    FILE *file = fopen(spline_file, "r");
    if (file != NULL) {
        int null = fscanf(file, "%1000000s", spline_knots);
	if (null != 1) {
	    cpl_msg_warning(cpl_func, "Error in parsing the spline knots.");  
	}
        fclose(file);
    }
    char *knots_repl = (char *)malloc(sizeof(char) * LEN_STRING); //[LEN_STRING];
    espda_dfs_ensure(espda_param_replace(spline_knots, ":", ",", knots_repl));
    cpl_array *spline_wavel = espda_param_extract(knots_repl, ",", 0, 2, 
        CPL_TYPE_DOUBLE);
    cpl_array *spline_flux = espda_param_extract(knots_repl, ",", 1, 2, 
        CPL_TYPE_DOUBLE);
    cpl_free(knots_repl);
    cpl_table *spline_man = NULL;
    if (spline_wavel != NULL && spline_flux != NULL) {
        cpl_size wavel_size = cpl_array_get_size(spline_wavel);
        cpl_size flux_size = cpl_array_get_size(spline_flux);
	if (wavel_size == flux_size) {
	    spline_man = cpl_table_new(wavel_size);
	    cpl_table_new_column(spline_man, "WAVEL", CPL_TYPE_DOUBLE);
	    cpl_table_new_column(spline_man, "FLUX", CPL_TYPE_DOUBLE);
	    cpl_table_new_column(spline_man, "CONT", CPL_TYPE_DOUBLE);
	    cpl_table_copy_data_double(spline_man, "WAVEL",
				       cpl_array_get_data_double(spline_wavel));
	    cpl_table_copy_data_double(spline_man, "FLUX",
				       cpl_array_get_data_double(spline_flux));
	    cpl_table_fill_column_window_double(spline_man, "CONT", 0,
						cpl_table_get_nrow(spline_man),
						INV_DOUBLE);
	}
	else {
  	    cpl_msg_warning(cpl_func, " Wavelength and flux values of spline "
			    "knots do not match in number! Ignoring them.");
	}
    }
    //espda_struct_sort(spline_man, 1, "WAVEL", FALSE);
    //cpl_table_dump(spline_man, 0, 1000, NULL);
    
/*
 * Redshift is loaded from the equalized spectrum also for the rebinned spectrum
 * and for the list of lines (it is assumed that the two spectra are of the same
 * object and that the lines are extracted from one of them).
 */

    double zem;
    espda_frame_desc(frameset, RSPEC, DAS_Z_EM, &zem); 
    
/*
 * The spectra (original and corrected) and the optional line list are split 
 * into parts, redwards and bluewards from the Lyman-alpha emission.
 */

    cpl_msg_info(cpl_func, "*** Split red and blue parts *** ");
    double lya_zem = (1.0 + zem) * (1 - 1500.0 / LIGHTSPEED) * WAVEL_LY_A;
    //double lya_zem = (1.0 + zem) * WAVEL_NV_1242;
    //double lya_zem = (1.0 + zem) * WAVEL_MGII_2803;
    //cpl_table *fspec_b = NULL, *fspec_r = NULL;
    cpl_table *rspec_b = NULL, *rspec_r = NULL;
    cpl_table *fline_b = NULL, *fline_r = NULL;
/**/
    //cpl_table *fprof_b = NULL, *fprof_r = NULL;
    cpl_table *rprof_b = NULL, *rprof_r = NULL;
/**/
    //espda_util_safe(fspec, espda_split_part(fspec, lya_zem, &fspec_b, &fspec_r)); 
#define MAX(X, Y) (((X) > (Y)) ? (X) : (Y))
    int fact = 1.01;
    //double edge = MAX(vel_sampl_blue, vel_sampl_red) * fact;
    //edge = 0.0;
    espda_dfs_call(0, espda_split_part(rspec, lya_zem, 0.0, &rspec_b, 
				       &rspec_r));
    espda_dfs_call(0, espda_split_part(fline_pre, lya_zem, 0.0, &fline_b, 
				       &fline_r)); 
/**/
    //espda_util_safe(fprof, espda_split_part(fprof, lya_zem, &fprof_b, 
    //    &fprof_r)); 
    espda_util_safe(rprof, espda_split_part(rprof, lya_zem, 2000, &rprof_b, 
        &rprof_r));

    cpl_table *spline_man_b = NULL, *spline_man_r = NULL;
    if (spline_man != NULL) {
        espda_util_safe(spline_man, espda_split_part(spline_man, lya_zem,
						     2000.0, &spline_man_b,
						     &spline_man_r));
    }
    //cpl_table_dump(spline_man_b, 0, 100, NULL);  
    //cpl_table_dump(spline_man_r, 0, 100, NULL);  
/**/
    
/*
 * The fitting procedure is initialized for the blue and the red parts:
 *
 *   - In the blue part, the continuum is fitted iteratively. The number of 
 *     iterations (controlled by fit_flag) corresponds to the total number of 
 *     lines that will be fitted and removed from the spectra. One line will be
 *     fitted at the first iteration, two at the second, etc. At each iteration
 *     the continuum is re-estimated by undersampling the spectrum and 
 *     interpolating it with a cubic spline. Prior to the interpolation, the 
 *     flux is enhanced by a corrective term which parametrizes the absorption
 *     of the lines that have not been fitted yet; this corrective term is 
 *     extracted from the column density distribution of the fitted lines. 
 *   
 *     NB: Lines are fitted w.r.t. two different continuum levels. To determine
 *         the column density distribution, the adopted continuum is the one  
 *         extracted from the enhanced flux (the "right" continuum, which is
 *         saved in the science spectra fspec_* and rspec_*). To remove the
 *         lines, the adopted continuum is the one extracted from the original
 *         flux (saved in the control spectrum remspec_*): otherwise, line 
 *         removal is biased.
 *
 *   - In the red part, the continuum is fitted at once by fitting and removing
 *     all the lines. 
 */

    int part_min = 1;
    int part_max = 1;
    if (rspec_b != NULL) {
        part_min = 0;
    }
    if (rspec_r  != NULL) {
        part_max = 2;
    }
    
    // Compute instrument profile    
    cpl_table *inst_prof = espda_model_inst(MODEL_SIZE, model_resol, 
        MODEL_KAPPA);
    cpl_table *rspec_prof = NULL;

    // Loop over parts
    cpl_table *fspec_cont     = NULL;
    cpl_table *fspec_rem      = NULL;
    cpl_table *rspec_cont_pre = NULL;
    cpl_table *rspec_rem_pre  = NULL;
    cpl_table *rspec_cont     = NULL;
    cpl_table *rspec_rem      = NULL;
    cpl_table *fline_rem      = NULL;

    double colden_thres = 100;
    if (spline_man == NULL) {
        colden_thres = 0;
    }  
    grp_mode = "OVER";

    cpl_table *spline_cont = NULL; 
    
    for (int part_i = part_min; part_i < part_max; part_i++) {
        
        // Setup part
        cpl_table *rspec_p = NULL;
        cpl_table *fline_p = NULL;
/**/
        cpl_table *rprof_p = NULL;
/**/
        double colden_old = 100; 
	if (spline_man == NULL) {
	    colden_old = 0;
	}  

        double spline_step, wavel_start = 0.0, wavel_end = 0.0;
        double piece_step = vel_sampl_red;
        int iter_max = 1;

        if (part_i == 0) {
            cpl_msg_info(cpl_func, "*** Process blue part ***");
            espda_struct_check(rspec_b, rspec_p = cpl_table_duplicate(rspec_b));
            espda_struct_check(fline_b, fline_p = cpl_table_duplicate(fline_b));
/**/
            espda_struct_check(rprof_b, rprof_p = cpl_table_duplicate(rprof_b));
/**/
            spline_step = vel_sampl_blue * fact;
            piece_step = vel_sampl_blue;
            espda_struct_check(rspec_p, espda_struct_startend(rspec_p, "WAVEL", 
                "WAVEL", wavel_end, wavel_start));
            espda_struct_size(fline_p, iter_max, 1);
        }
        else {
            cpl_msg_info(cpl_func, "*** Process red  part ***");
            espda_struct_check(rspec_r, rspec_p = cpl_table_duplicate(rspec_r));
            espda_struct_check(fline_r, fline_p = cpl_table_duplicate(fline_r));
/**/
            espda_struct_check(rprof_r, rprof_p = cpl_table_duplicate(rprof_r));
/**/
            spline_step = vel_sampl_red * fact;
            piece_step = vel_sampl_red;
            espda_struct_check(rspec_p, espda_struct_startend(rspec_p, "WAVEL", 
                "WAVEL", wavel_start, wavel_end));

            //espda_struct_size(fline_p, iter_max, 1); // Treat red part like blue

        }
        if (fline_p == NULL) {
            cpl_msg_info(cpl_func, " No lines to fit!");
	    continue;
        }
	    
	
        // Limit the number of iterations
        if (iter_max > iter_max_temp && iter_max_temp != -1) {
            iter_max = iter_max_temp;   
        }

	// Skip iterations if spline knots are provided
	cpl_table *spline_man_p = NULL;
	if (spline_man != NULL) {
            cpl_msg_info(cpl_func, " Using manually-entered knots.");
	    iter_max = 1;
	    if (part_i == 0) {
	        spline_man_p = cpl_table_duplicate(spline_man_b);
		espda_struct_sort(spline_man_p, 1, "WAVEL", TRUE);
	    }
	    else {
	        spline_man_p = cpl_table_duplicate(spline_man_r);
	    }
	}
        cpl_table *rspec_cont_in = NULL; 
        cpl_table *rspec_rem_in  = NULL;
        //cpl_table *rspec_tails   = NULL; 
        espda_struct_check(rspec_p, rspec_cont_in = 
            cpl_table_duplicate(rspec_p));
        espda_struct_check(rspec_p, rspec_rem_in  = 
            cpl_table_duplicate(rspec_p));
       //iter_max = 1;
        for (int iter_i = 0; iter_i < iter_max; iter_i++) {

/*
 * Reference and guess continua are computed, after correcting the flux for 
 * Lyman-alpha absorption. Reference continuum is the one used to define line 
 * peaks; it is computed by dramatically undersampling the whole spectrum and 
 * fitting it with a cubic spline. Guess continuum is the one to be refined by
 * incremental line removal; it is computed by undersampling the blue or the red
 * part at the velocity step defined by the user and fitting it with a cubic 
 * spline.
 *
 * NB: Determination of the reference continuum will be improved, using a 
 *     power-law in selected regions of the spectrum.
 */

            // Check if it is the last iteration
            cpl_table *temp = NULL;
            espda_struct_check(fline_p, espda_struct_extract_old(fline_p, temp, 
                "COLDEN", CPL_TYPE_DOUBLE, CPL_EQUAL_TO, INV_DOUBLE));
            cpl_size iter_check = 0;
            espda_struct_check(temp, iter_check = cpl_table_get_nrow(temp));
            cpl_table_delete(temp);
            if (iter_check <= 1 && iter_i > 0) {
                iter_max = iter_i + 1;
            }
    
            // Enhance flux in spectra
            cpl_table *rspec_corr     = NULL;
            cpl_table *rspec_rem_corr = NULL;
            espda_dfs_call(0, espda_spec_corr(rspec_p,      zem, colden_thres, 
                1, &rspec_corr));

            espda_dfs_call(0, espda_spec_corr(rspec_rem_in, zem, colden_thres, 
                1, &rspec_rem_corr));

            // Rebin spectra with lines removed to define piece and spline knots
            cpl_table *spline = NULL;
            cpl_table *pieces = NULL;

	    /*** COMMENT HERE TO USE SAVITZKY-GOLAY ***/
	    //cpl_table_dump(rspec_rem_corr, 0, 10, NULL);
            espda_util_func(spline = espda_spec_rebin(rspec_rem_corr,  3, 
                wavel_start, wavel_end, spline_step), SPEC_REBIN, 0);
	    /***/
	    //cpl_table_dump(spline, 0, 1000, NULL);
	    /*printf("%f, %f, %f, %f\n",
		   cpl_table_get_column_min(rspec_rem_corr, "WAVEL"),
		   cpl_table_get_column_max(rspec_rem_corr, "WAVEL"),
		   cpl_table_get_column_min(spline, "WAVEL"),
		   cpl_table_get_column_max(spline, "WAVEL"));
	    */
	    
	    // Correct spline with manually-entered knots, if provided
	    double spline_start = INV_DOUBLE, spline_end = INV_DOUBLE;
	    if (part_i == 0) {
	        espda_struct_startend(spline, "WAVEL", "WAVEL", spline_end,
				      spline_start);
	    }
	    else {
	        espda_struct_startend(spline, "WAVEL", "WAVEL", spline_start,
				      spline_end);
	    }
	    //cpl_table_dump(spline, 0, 100, NULL);
	    //printf("%f %f\n", spline_start, spline_end);
	    
	    espda_struct_erase(spline_man_p, 2, "WAVEL", CPL_LESS_THAN,
			       spline_start-1e-6, "WAVEL", CPL_GREATER_THAN,
			       spline_end+1e-6);
	    //cpl_table_dump(spline_man_p, 0, 100, NULL);
	    //printf("%i %i\n",  cpl_table_get_nrow(spline), cpl_table_get_nrow(spline_man_p));
	    
	    /*
	    if (spline_man_p != NULL && cpl_table_get_nrow(spline) == cpl_table_get_nrow(spline_man_p)) {
	    */
	    ///*
	    if (spline_man_p != NULL) { 
	        if (cpl_table_get_nrow(spline)
		    < cpl_table_get_nrow(spline_man_p)) {
  		    cpl_table_insert_window(spline, cpl_table_get_nrow(spline),
		  		            cpl_table_get_nrow(spline_man_p)
				            - cpl_table_get_nrow(spline));
		}
	    //*/
		cpl_table_copy_data_double(spline, "WAVEL",
		    cpl_table_get_data_double(spline_man_p, "WAVEL"));
	        cpl_table_copy_data_double(spline, "FLUX",
		    cpl_table_get_data_double(spline_man_p, "FLUX"));
		espda_struct_sort(spline, 1, "WAVEL", FALSE);
	    }
	    /*
	    if (spline_man_p != NULL) {
	        spline = cpl_table_duplicate(spline_man_p);
	    }
	    */

	    //cpl_table_dump(spline, 0, 1000, NULL);
	    
	    // Add a first and last point to the spline

	    cpl_table *first_spline = cpl_table_extract(spline, 0, 1);
	    cpl_table *last_spline = cpl_table_extract(
		spline, cpl_table_get_nrow(spline)-1, 1);
	    double min = cpl_table_get_column_min(rspec_rem_corr, "WAVEL")
                         +cpl_table_get_double(rspec_rem_corr, "PIXSIZE", 0,
					       NULL);
	    double max = cpl_table_get_column_max(rspec_rem_corr, "WAVEL")
                         -cpl_table_get_double(
			      rspec_rem_corr, "PIXSIZE",
			      cpl_table_get_nrow(rspec_rem_corr)-1, NULL);
            //printf("%f %f %f\n", min, cpl_table_get_column_min(rspec_rem_corr, "WAVEL"), cpl_table_get_double(rspec_rem_corr, "PIXSIZE", 0,
	    //				       NULL));


	    if (part_i == 0) {
		if (spline_start > min) {
		    cpl_table_set_double(first_spline, "WAVEL", 0, min);
		}
		if (spline_end < max) {
		    cpl_table_set_double(last_spline, "WAVEL", 0, max);
		}
	    }
	    if (part_i == 1) {
		if (spline_start > min) {
		    cpl_table_set_double(first_spline, "WAVEL", 0, min);
		}
		if (spline_end < max) {
		    cpl_table_set_double(last_spline, "WAVEL", 0, max);
		}
	    }
	    if (part_i == part_max-1) {
		if (spline_end < max) {
		    cpl_table_insert(spline, last_spline,
		  		     cpl_table_get_nrow(spline));
		}
	    }

            //printf("%i %i %f %f\n", part_i, part_min, spline_start, min);
	    //cpl_table_dump(spline, 0, 200, NULL);
	    if (part_i == part_min) {
		if (spline_start > min) {
		    cpl_table_insert(spline, first_spline, 0);
		}
	    }
	    //cpl_table_dump(spline, 0, 200, NULL);

	    // Save spline at last iteration
            if (iter_i == iter_max - 1) {
		if (part_i == part_min) {
		    spline_cont = cpl_table_duplicate(spline);
		}
		else {
		    cpl_table_insert(spline_cont, spline,
				     cpl_table_get_nrow(spline_cont));
		}
            }
	    
            espda_util_func(pieces = espda_spec_rebin(rspec_rem_corr,  3, 
                wavel_start, wavel_end, piece_step),  SPEC_REBIN, 0);
	    espda_struct_sort(pieces, 1, "WAVEL", FALSE);
	    
	    // Add a first and last point to the pieces
	    double pieces_start = INV_DOUBLE, pieces_end = INV_DOUBLE;
	    espda_struct_startend(pieces, "WAVEL", "WAVEL", pieces_start,
				  pieces_end);

	    cpl_table *first_pieces = cpl_table_extract(pieces, 0, 1);
	    cpl_table *last_pieces = cpl_table_extract(
		pieces, cpl_table_get_nrow(pieces)-1, 1);
	    if (pieces_start > min) {
	        cpl_table_set_double(first_pieces, "WAVEL", 0, min);
	    }
	    if (pieces_end < max) {
	        cpl_table_set_double(last_pieces, "WAVEL", 0, max);
	    }
	    if (part_i == part_max-1) {
		if (pieces_end < max) {
		    cpl_table_insert(pieces, last_pieces,
				     cpl_table_get_nrow(pieces));
		}
	    }
	    if (part_i == part_min) {
		if (pieces_start > min) {
		    cpl_table_insert(pieces, first_pieces, 0);
		}
	    }
	    
	    double pieces_wavel_min = cpl_table_get_double(
		pieces, "WAVEL", 0,NULL);
	    double pieces_wavel_max = cpl_table_get_double(
		pieces, "WAVEL",  cpl_table_get_nrow(pieces)-1, NULL);
	    
	    cpl_table *rspec_rem_corr_temp = NULL;
	    espda_struct_window(rspec_rem_corr, "WAVEL", pieces_wavel_min, 
				pieces_wavel_max, "IN", rspec_rem_corr_temp);

	    // Fit spline knots to extract the reference continuum
	    
	    /*** COMMENT HERE TO USE SAVITZKY-GOLAY ***/	    
            //spline = espda_fit_sg(rspec_rem_corr_temp, "FLUX", sg_width_pieces);
	    /***/
	    //pieces = espda_fit_sg(rspec_rem_corr, "CONT", 200);
	    //cpl_table_save(pieces, NULL, NULL, "pieces.fits", CPL_IO_CREATE);
	    //cpl_table_save(spline, NULL, NULL, "spline.fits", CPL_IO_CREATE);
	    
            espda_struct_check(spline, espda_struct_sort_old(spline, "WAVEL", 
                FALSE));
            espda_struct_check(pieces, espda_struct_sort_old(pieces, "WAVEL", 
                FALSE));
	    //cpl_table_dump(spline, 0, 100, NULL);
	    espda_util_safe(rspec_p,        espda_util_func(espda_fit_cspline(
                spline, "CONT", rspec_p,        fline_p), FIT_CSPLINE, 0));  
            espda_util_safe(rspec_corr,     espda_util_func(espda_fit_cspline(
                spline, "CONT", rspec_corr,     NULL),    FIT_CSPLINE, 0));
            espda_util_safe(rspec_rem_corr, espda_util_func(espda_fit_cspline(
                spline, "CONT", rspec_rem_corr, NULL),    FIT_CSPLINE, 0)); 
	    if (iter_i == 0) {
	        cpl_msg_info(cpl_func, "%i lines will be fitted and "
			     "removed in at most %i iterations.",
			     (int)cpl_table_get_nrow(fline_p), iter_max);
	    }
	    /*
	    printf("%f, %f, %f, %f\n",
		   cpl_table_get_column_min(rspec_p, "WAVEL"),
		   cpl_table_get_column_max(rspec_p, "WAVEL"),
		   cpl_table_get_column_min(spline, "WAVEL"),
		   cpl_table_get_column_max(spline, "WAVEL"));
	    */
	    cpl_table_delete(spline);
            int piece_max;
            espda_struct_size(pieces, piece_max, 1);
            piece_max--;

            // Interpolate instrument profile 
            // N.B. do not move this block
            if (rprof == NULL && iter_i == 0) {
                espda_dfs_call(0, espda_interp_inst(rspec_p, inst_prof, 
                    MODEL_SIZE, model_resol, MODEL_KAPPA, &rspec_prof));
            }    
 
            // Prepare line list for fitting
            cpl_table *fline_full = NULL;
            if (fline_p != NULL) {
                cpl_table *fline_void = NULL;
                if (iter_i != 0) {
                    espda_struct_extract_old(fline_p, fline_void, "COLDEN", 
                        CPL_TYPE_DOUBLE, CPL_EQUAL_TO, INV_DOUBLE);                   
                    espda_struct_extract_old(fline_p, fline_full, "COLDEN", 
                        CPL_TYPE_DOUBLE, CPL_NOT_EQUAL_TO, INV_DOUBLE);
                }   
                else {
                    espda_struct_check(fline_p, fline_void = 
                        cpl_table_duplicate(fline_p));
                }
                if (fline_void != NULL) {
                    cpl_table *fline_new = NULL;
                    espda_dfs_call(0, espda_voigt_prep(fline_void, ion, hwidth, 
                        zem, par_range, &fline_new));
                    colden_old = colden_thres;
                    int thres;
                    espda_struct_size(fline_full, thres, 0);
                    thres = (int)round(0.75 * thres);
                    if (thres != 0) {
                        espda_struct_check(fline_full, 
                            espda_struct_sort_old(fline_full, "COLDEN", TRUE));
                        espda_struct_check(fline_full, colden_thres =
                            cpl_table_get_double(fline_full, "COLDEN", thres, 
                            NULL));
                    }
                    espda_struct_insert(fline_full, fline_new, 
                        cpl_table_get_nrow(fline_full));
                }
            }

	    // At the last iteration, add fitting residuals to the line list
	    if (iter_i == iter_max - 1 && 1==2) {
		int jmax = 5;
		cpl_table *fline_res = NULL;
		for (int j = 1; j < jmax + 1; j++) {
  		    double fact = j/(double)jmax;
		    double vel_tmp = 20.0 * fact;
		    cpl_table *fline_tmp = NULL;
		    cpl_table *rspec_smooth = espda_spec_rebin(rspec_rem_corr,
							       3, wavel_end,
							       wavel_start,
							       vel_tmp);
		    espda_util_func(fline_tmp = espda_line_detect(rspec_smooth,
								  rspec_smooth,
								  3.0),
				    LINE_DETECT, 0);
		    if (j == 1) {
		        fline_res = cpl_table_duplicate(fline_tmp);
		    }
		    else {
 		        cpl_table_insert(fline_res, fline_tmp, 0);
		    }
		    espda_util_safe(fline_res, espda_util_func(espda_line_clean(
		        fline_res, vel_tmp*0.5), LINE_CLEAN, 0));
		}
		
		cpl_table *fline_res_form = NULL;
		espda_dfs_call(0, espda_voigt_prep(fline_res, ion, hwidth, 
                        zem, par_range, &fline_res_form));
		cpl_table_insert(fline_full, fline_res_form, 0);
		espda_util_safe(fline_full, espda_util_func(espda_line_clean(
		    fline_full, 0.1), LINE_CLEAN, 0));
		//cpl_table_save(fline_res_form, NULL, NULL, "fline_res_form.fits", CPL_IO_CREATE);
		//cpl_table_save(rspec_rem_corr, NULL, NULL, "rspec_rem_corr.fits", CPL_IO_CREATE);
		//cpl_table_dump(fline_res_form, 0, 10, NULL);

	    }
	    if (iter_i == iter_max - 1) {
	      espda_struct_sort(fline_full, 1, "WAVEL", FALSE);
	      //cpl_table_dump(fline_full, 0, 20, NULL);
	    }

	    
/*
 * Pieces are initialized for the blue and the red parts.
 */
                    
            cpl_table *rspec_cont_piece = NULL;
            cpl_table *rspec_rem_piece = NULL;            
            cpl_table *fline_piece     = NULL;
            //cpl_size   fline_piece_num;
            espda_struct_check(rspec_corr,     rspec_cont_piece =
                cpl_table_duplicate(rspec_corr));
            espda_struct_check(rspec_rem_corr, rspec_rem_piece  = 
                cpl_table_duplicate(rspec_rem_corr));
	    
            cpl_table_delete(rspec_rem_corr);
            espda_struct_check(fline_full, espda_struct_initcol(fline_full, 
                "GRPID", INV_INT));
            espda_dfs_call(0, espda_voigt_prep(fline_full, ion, hwidth, zem,
                par_range, &fline_piece));
            cpl_table_delete(fline_full);
            //espda_struct_size(fline_piece, fline_piece_num, 0);
            espda_struct_check(fline_piece, espda_struct_sort_old(fline_piece, 
                "WAVEL", FALSE));

            int fline_sel_max = 0, fline_sel_tot = 0, iter_stop = 0;
            if (iter_i == iter_max - 1) {
                espda_print_plurt(fline_full,   fline_full_plur);
                espda_print_plurn(piece_max, piece_max_plur);
                cpl_msg_info(cpl_func, " Last iteration: fitting and removing "
                    "%3i line%s in %2i piece%s...", 
                    (int)cpl_table_get_nrow(fline_full), 
                    fline_full_plur, piece_max, piece_max_plur);
            }
        
            // Loop over pieces
            /* AMO: NB IF YOU PUT OMP STUFF BEWARE OSX10/11 DO NOT SUPPORT IT
               ==> you need to control OMP STATEMENTS BY
              #ifdef _OPENMP
                ...<your omp code>...
              #endif
            */
	    //#pragma omp parallel for //num_threads(3)
            for (int piece_i = 0; piece_i < piece_max; piece_i++) {
	      

	        // Define piece
                double piece_wavel_min = cpl_table_get_double(pieces, "WAVEL", 
                    piece_i, NULL);
                double piece_wavel_max = cpl_table_get_double(pieces, "WAVEL", 
                    piece_i + cpl_table_get_nrow(pieces) - piece_max, NULL);
                int prof_pos_min = INV_INT, prof_pos_max = INV_INT;
                cpl_table *rspec_prof_temp = NULL;

		if (rprof == NULL) {
                    espda_struct_window(rspec_prof, "WAVEL", piece_wavel_min, 
                        piece_wavel_max, "IN", rspec_prof_temp);
		    if (rspec_prof_temp != NULL) {
                        espda_struct_startend(rspec_prof_temp, "PROFMIN", 
			    "PROFMAX", prof_pos_min, prof_pos_max);
		    }
		    else {
		        continue;
		    }
		}

/**/

		// Define the instrument profile to select the chunk
//		if (rprof != NULL) {
		else {
                    // The following lines are needed to adapt the general 
                    // instrument profile to the selected chunk
		    cpl_size rprof_diff = cpl_table_get_int(rprof_p, "PROFMAX",
                        0, NULL);
                    cpl_size rprof_shift = (int)ceil(0.5 * (rprof_diff -
		        cpl_table_get_int(rprof_p, "PROFMIN", 0, NULL))); 
		    cpl_size rprof_size = cpl_table_get_nrow(rprof_p);
                    cpl_table_subtract_scalar(rprof_p, "PROFMIN", rprof_diff -
                        rprof_shift);	
                    cpl_table_subtract_scalar(rprof_p, "PROFMAX", rprof_diff -
                        rprof_shift);

		    for (int p = 0; p < rprof_shift; p++) {
		        cpl_table_set_int(rprof_p, "PROFMIN", p, 0);
			cpl_array *head = NULL;
		        const cpl_array *htmp =
			  cpl_table_get_array(rprof_p, "PROFNORM", p);
			cpl_array_copy_data_double(
			    head, cpl_array_get_data_double_const(htmp));
                        cpl_array_erase_window(head, 0, rprof_shift - 1 - p);
		        cpl_array_set_size(head, rprof_shift * 2 + 1);
		        cpl_table_set_array(rprof_p, "PROFNORM", p, head);
		        cpl_table_set_int(rprof_p, "PROFMAX", rprof_size - 1 -
                            p, cpl_table_get_int(rprof_p, "PROFMAX", rprof_size
                            - rprof_shift, NULL));

		        cpl_array *tail = NULL;
		        const cpl_array *ttmp =
			  cpl_table_get_array(rprof_p, "PROFNORM",
					      rprof_size - p);
			cpl_array_copy_data_double(
			    tail, cpl_array_get_data_double_const(ttmp)); 
                        cpl_array_erase_window(tail, rprof_shift + p, 
                            rprof_shift * 2);
		        cpl_array_insert_window(tail, 0, rprof_shift - p);
		        cpl_table_set_array(rprof_p, "PROFNORM", rprof_size - p,
                            tail);
                    }

                    cpl_table *rprof_temp = NULL;
                    espda_struct_window(rprof_p, "WAVEL", piece_wavel_min, 
                        piece_wavel_max, "IN", rprof_temp);
                    espda_struct_check(rprof_temp, 
                        espda_struct_startend(rprof_temp, "PROFMIN", "PROFMAX",
                        prof_pos_min, prof_pos_max));
                    cpl_table_delete(rprof_temp);
		}   
                cpl_table_delete(rspec_prof_temp);

/**/
                cpl_table *rspec_sel       = NULL;
                espda_struct_extract_old_2(rspec_corr, rspec_sel,  
                    prof_pos_min, prof_pos_max - prof_pos_min);
                
/**/

                cpl_table *rprof_sel       = NULL;
		    if (rprof != NULL) {
                    espda_struct_extract_old_2(rprof_p, rprof_sel,  
                        prof_pos_min, prof_pos_max - prof_pos_min);
                    // The following lines are needed to adapt the general 
                    // instrument profile to the selected chunk
		    cpl_size rprof_diff = cpl_table_get_int(rprof_sel, "PROFMAX",
                        0, NULL);
                    cpl_size rprof_shift = (int)ceil(0.5 * (rprof_diff -
		        cpl_table_get_int(rprof_sel, "PROFMIN", 0, NULL))); 
		    cpl_size rprof_size = cpl_table_get_nrow(rprof_sel);
                    /*
		    rprof_diff = cpl_table_get_int(rprof_sel, "PROFMAX",
                        0, NULL);
                    rprof_shift = (int)ceil(0.5 * (rprof_diff -
		        cpl_table_get_int(rprof_sel, "PROFMIN", 0, NULL))); 
		    rprof_size = cpl_table_get_nrow(rprof_sel);
		    */
		    cpl_table_subtract_scalar(rprof_sel, "PROFMIN", rprof_diff -
                        rprof_shift);	
                    cpl_table_subtract_scalar(rprof_sel, "PROFMAX", rprof_diff -
                        rprof_shift);

		    for (int p = 0; p < rprof_shift; p++) {
		        cpl_table_set_int(rprof_sel, "PROFMIN", p, 0);
		        cpl_array *head = NULL;
		        const cpl_array *htmp =
			  cpl_table_get_array(rprof_sel, "PROFNORM", p);
			cpl_array_copy_data_double(
			    head, cpl_array_get_data_double_const(htmp));
                        cpl_array_erase_window(head, 0, rprof_shift - 1 - p);
		        cpl_array_set_size(head, rprof_shift * 2 + 1);
		        cpl_table_set_array(rprof_sel, "PROFNORM", p, head);
		        cpl_table_set_int(rprof_sel, "PROFMAX",
					  rprof_size - 1 - p,
		    	cpl_table_get_int(rprof_sel, "PROFMAX", rprof_size
                            - rprof_shift, NULL));
			
		        cpl_array *tail = NULL;
		        const cpl_array *ttmp =
			  cpl_table_get_array(rprof_sel, "PROFNORM",
					      rprof_size - p);
			cpl_array_copy_data_double(
			    tail, cpl_array_get_data_double_const(ttmp));
                        cpl_array_erase_window(tail, rprof_shift + p, 
					       rprof_shift * 2);
		        cpl_array_insert_window(tail, 0, rprof_shift - p);
		        cpl_table_set_array(rprof_sel, "PROFNORM",
					    rprof_size - p, tail);
                    }   
                }

/**/

                // Extract piece from line list
                cpl_table *fline_sel = NULL;
                cpl_table *fline_rej = NULL;
                espda_struct_window(fline_piece, "WAVEL", piece_wavel_min, 
                    piece_wavel_max, "IN",  fline_sel);
                espda_struct_window(fline_piece, "WAVEL", piece_wavel_min, 
                    piece_wavel_max, "OUT", fline_rej);
                cpl_size fline_sel_num;
                espda_struct_size(fline_sel, fline_sel_num, 0);
                if (fline_sel_num > fline_sel_max) {
                    fline_sel_max = fline_sel_num;
                }

/*
 * Lines are incrementally fitted. The threshold column density is computed as 
 * the median of the column density distribution.
 */

                // Select lines to process and to discard
                cpl_table *fline_proc = NULL;
                cpl_table *fline_disc = NULL;
                if (part_i == 0) {
		  //if (part_i >= 0) {
                    espda_struct_check(fline_sel, espda_struct_sort_old(
                        fline_sel, "COLDEN", TRUE));
                }

                // To fit all lines at the last iteration
                iter_stop = fline_sel_num;
                if (iter_stop > iter_max) {
                    iter_stop = iter_max;
                }
                if (part_i == 0 && iter_i < iter_stop - 1) {  
                  //if (part_i >= 0 && iter_i < iter_stop - 1) {  
                    fline_proc = cpl_table_extract(fline_sel, 0, iter_i + 1); 
                    
                    // Parameters of distant lines from other are fixed
                    double wavel_new = cpl_table_get_double(fline_proc, 
                        "WAVEL", iter_i, NULL);
                    for (int line_i = 0; line_i < iter_i; line_i++) {
                        double peak = cpl_table_get_double(fline_proc, 
                            "PEAK", line_i, NULL);
                        double cont = cpl_table_get_double(fline_proc, 
                            "CONT", line_i, NULL);
                        double wavel = cpl_table_get_double(fline_proc, 
                            "WAVEL", line_i, NULL);
                        double colden = cpl_table_get_double(fline_proc, 
                            "COLDEN", line_i, NULL);
                        double therbb = cpl_table_get_double(fline_proc, 
                            "THERB", line_i, NULL);
                        double  colden_ref = espda_model_colden(peak, cont);
                        double  therb_ref  = 20.0;
                        if (wavel > WAVEL_LY_A * (1.0 + zem)) {
                            colden_ref = colden_ref - 0.5;
                            therb_ref  = 10.0;
                        }
                        if (fabs(wavel - wavel_new) > 2 * hwidth &&
                            colden != colden_ref && therbb != therb_ref) {                        
                            //cpl_table_set_string(fline_proc, "LINEZCON", 
                            //    line_i, "F");
                            cpl_table_set_string(fline_proc, "COLDENCON", 
                                line_i, "F");
                            cpl_table_set_string(fline_proc, "THERBCON", 
                                line_i, "F");
                            cpl_table_set_string(fline_proc, "TURBBCON", 
                                line_i, "F");
                        }
                    }
                    
                    fline_disc = cpl_table_extract(fline_sel, iter_i + 1, 
                        fline_sel_num);
                    cpl_table_fill_column_window_double(fline_disc, "COLDEN", 
                        0, cpl_table_get_nrow(fline_disc), INV_DOUBLE);
                    cpl_table_insert(fline_rej, fline_disc, 
                        cpl_table_get_nrow(fline_rej));
                }
                else {
                    espda_struct_check(fline_sel, fline_proc = 
                        cpl_table_duplicate(fline_sel)); 
                    /*for (int line_i = 0; line_i < 
			   cpl_table_get_nrow(fline_proc); line_i++) {
                        double peak = cpl_table_get_double(fline_proc, 
                            "PEAK", line_i, NULL);
                        double cont = cpl_table_get_double(fline_proc, 
                            "CONT", line_i, NULL);
                        double  colden_ref = espda_model_colden(peak, cont) - 
			    1.0;
                        double  therb_ref  = 5.0;
			cpl_table_set_double(fline_proc, "COLDEN", line_i, 
					     colden_ref);
			cpl_table_set_double(fline_proc, "THERB", line_i, 
					     therb_ref);
					     }*/
                }
		//cpl_table_dump(fline_proc, 0, 10, NULL);
                cpl_table_delete(fline_sel);
                cpl_table_delete(fline_disc);
     
                // Fit and remove lines
                cpl_table *rspec_cont_temp = NULL;
                cpl_table *rspec_rem_temp = NULL;
                cpl_table *fline_fit_temp = NULL;
                cpl_size fline_proc_num;
                espda_struct_size(fline_proc, fline_proc_num,  0);
                fline_sel_tot = fline_sel_tot + fline_proc_num;
                espda_print_plurt(fline_proc, fline_proc_plur);
                cpl_array *rchisq_fit = NULL;
                cpl_array *iter_fit   = NULL;                
                cpl_table *null_sel = NULL, *null_fit_temp = NULL;

		//cpl_table *fprof = NULL;
		cpl_table *rprof = NULL;
/*
                espda_dfs_call(0, espda_wrap_line(null_sel, rspec_sel, 
                    fline_proc, ion, voigt_func, model_resol, 1000, 0, 1, 
                    fprof, rprof, &null_fit_temp, &rspec_cont_temp,
		    &fline_fit_temp, &rchisq_fit, &iter_fit));
*/
/**/
		//cpl_table_dump(fline_proc, 0, 10, NULL);
		//printf("%i\n", (int)cpl_table_get_nrow(fline_proc));
                espda_dfs_call(0, espda_wrap_line(null_sel, rspec_sel, 
                    fline_proc, ion, voigt_func, model_resol, 10, 0, 1, 
                    NULL, rprof_sel, &null_fit_temp, &rspec_cont_temp,
		    &fline_fit_temp, &rchisq_fit, &iter_fit));


/**/
                //cpl_table_delete(fprof);
		cpl_table_delete(rprof);
                cpl_table_delete(rspec_sel);
                cpl_table_delete(fline_proc);                
                cpl_array_delete(rchisq_fit);
                cpl_array_delete(iter_fit);               
                //espda_util_safe(rspec_cont_temp, espda_util_func(
                //    rspec_rem_temp = espda_spec_remove(rspec_cont_temp, 
                //    peak_thres), SPEC_REMOVE, 0)); 
		espda_util_safe(rspec_cont_temp,
				rspec_rem_temp = espda_spec_remove(rspec_cont_temp, 
                    peak_thres));
		if (iter_i == iter_max - 1 && fline_sel_num > 0) {
                    cpl_msg_info(cpl_func, "  Piece %2i of %2i (wavelength "
                        "range [%6.2f,%6.2f]): fitted and removed %2i of %2i "
                        "line%s.", piece_i + 1, piece_max, piece_wavel_min, 
                        piece_wavel_max, (int)fline_proc_num, 
                        (int)fline_sel_num, fline_proc_plur);
                }
                
                // Save piece
                if (piece_i == 0) {
                    rspec_cont_piece = NULL;
                    rspec_rem_piece = NULL;
                }
                espda_struct_insert(rspec_cont_piece, rspec_cont_temp, 
                    cpl_table_get_nrow(rspec_cont_piece));
                espda_struct_insert(rspec_rem_piece, rspec_rem_temp, 
                    cpl_table_get_nrow(rspec_rem_piece));
                fline_piece = NULL;
                espda_struct_check(fline_fit_temp, fline_piece = 
                    cpl_table_duplicate(fline_fit_temp));
                espda_struct_insert(fline_piece, fline_rej, 
                    cpl_table_get_nrow(fline_piece));
                cpl_table_delete(fline_rej);
		/*printf("%f, %f\n",
		   cpl_table_get_column_min(rspec_cont_temp, "WAVEL"),
		   cpl_table_get_column_max(rspec_cont_temp, "WAVEL"));
		*/
		cpl_table_delete(rspec_cont_temp);
                cpl_table_delete(rspec_rem_temp);
                cpl_table_delete(fline_fit_temp);
            }
	    
//} // OPENMP brace

            // Diminish flux in spectra
	    cpl_table *rspec_fit_piece = cpl_table_duplicate(rspec_cont_piece);
	    cpl_table *rspec_fit_in = NULL;
	    cpl_table_copy_data_double(rspec_fit_piece, "FLUX",
		cpl_table_get_data_double(rspec_fit_piece, "FIT"));
            espda_dfs_call(0, espda_spec_corr(rspec_fit_piece, zem, colden_old,
                0, &rspec_fit_in));
	    cpl_table_copy_data_double(rspec_cont_piece, "FIT",
		cpl_table_get_data_double(rspec_fit_in, "FLUX"));	    

            espda_dfs_call(0, espda_spec_corr(rspec_cont_piece, zem, colden_old,
                0, &rspec_cont_in));
            espda_dfs_call(0, espda_spec_corr(rspec_rem_piece, zem, colden_old, 
                0, &rspec_rem_in));
            espda_struct_check(fline_piece, fline_p = 
                cpl_table_duplicate(fline_piece));
            cpl_table_delete(rspec_corr);
            cpl_table_delete(rspec_cont_piece);
            cpl_table_delete(rspec_rem_piece);
            cpl_table_delete(fline_piece);

            if (iter_i < iter_max - 1) {
                espda_print_plurn(fline_sel_tot, fline_sel_tot_plur);           
                espda_print_plurn(piece_max,     piece_max_plur);
                cpl_msg_info(cpl_func, " Iteration %2i:   fitted  and "
                    "removed  %3i line%s in %2i piece%s...", iter_i + 1, 
                    fline_sel_tot, fline_sel_tot_plur, piece_max, 
                    piece_max_plur);
            }           
        }
	//cpl_table_save(spline_cont, NULL, NULL, "spline_cont.fits", CPL_IO_CREATE);        
        // Enhance flux in spectra
        cpl_table *rspec_cont_out = NULL;
        cpl_table *rspec_rem_out = NULL;
        espda_struct_check(rspec_cont_in, rspec_cont_out =
            cpl_table_duplicate(rspec_cont_in));
        espda_dfs_call(0, espda_spec_corr(rspec_rem_in, zem, colden_old, 1,                                           
            &rspec_rem_out));
        cpl_table_delete(rspec_cont_in);
        cpl_table_delete(rspec_rem_in);
            
        // Clip the parts to avoid boundary issues
        cpl_table *rspec_cont_clip = NULL;
        cpl_table *rspec_rem_clip = NULL;
        if (part_i == 0) {
            rspec_cont_clip = espda_struct_extract(rspec_cont_out, 1, "WAVEL",
                CPL_NOT_GREATER_THAN, lya_zem);
            rspec_rem_clip  = espda_struct_extract(rspec_rem_out,  1, "WAVEL",
                CPL_NOT_GREATER_THAN, lya_zem);
        }
        if (part_i == 1) {
            rspec_cont_clip = espda_struct_extract(rspec_cont_out, 1, "WAVEL",
                CPL_GREATER_THAN, lya_zem);
            rspec_rem_clip  = espda_struct_extract(rspec_rem_out,  1, "WAVEL",
                CPL_GREATER_THAN, lya_zem);
        }
                
        // Save parts
        espda_struct_insert(rspec_cont_pre, rspec_cont_clip, 
            cpl_table_get_nrow(rspec_cont_pre));
	
        espda_struct_insert(rspec_rem_pre,  rspec_rem_clip, 
            cpl_table_get_nrow(rspec_rem_pre));
        espda_struct_insert(fline_rem, fline_p, cpl_table_get_nrow(fline_rem));
        cpl_table_delete(rspec_p);
        cpl_table_delete(rspec_cont_out);
        cpl_table_delete(rspec_rem_out);
        cpl_table_delete(rspec_cont_clip);
        cpl_table_delete(rspec_rem_clip);
    }
    espda_struct_delete(CPL_TABLE, 2, rspec_b, rspec_r);
    espda_struct_delete(CPL_TABLE, 2, rprof_b, rprof_r);

    espda_struct_check(fspec, fspec_cont = cpl_table_duplicate(fspec));
    espda_struct_check(fspec, fspec_rem  = cpl_table_duplicate(fspec));
    espda_struct_check(rspec, rspec_cont = cpl_table_duplicate(rspec));
    espda_struct_check(rspec_rem_pre, rspec_rem = 
        cpl_table_duplicate(rspec_rem_pre));
    if (fline_rem != NULL) {
        for (int i = 0; i < cpl_table_get_nrow(fline_rem); i++) {
            double wavel = cpl_table_get_double(fline_rem, "WAVEL", i, 
                NULL);
            double hwidth = 0.5 * cpl_table_get_column_mean(rspec_cont_pre, 
                "PIXSIZE");
            cpl_table *rspec_line = NULL;
            espda_struct_window(rspec_cont_pre, "WAVEL", wavel - hwidth, 
                wavel + hwidth, "IN", rspec_line);
            //cpl_table_set_double(fline_rem, "PEAK", i, 
            //    cpl_table_get_double(rspec_line, "FIT",  0, NULL) * 
            //    cpl_table_get_double(rspec_line, "CONT", 0, NULL));
        }
    }
    //double wavel_start = 0.0, wavel_end = 0.0;
    cpl_table *rspec_temp = NULL;
    espda_struct_check(rspec_cont_pre, rspec_temp = 
        cpl_table_duplicate(rspec_cont_pre));



    for (int i = 1; i < cpl_table_get_nrow(rspec_temp); i++) {
        if (cpl_table_get_double(rspec_temp, "WAVEL", i, NULL) 
            == cpl_table_get_double(rspec_temp, "WAVEL", i - 1, NULL)) {
            //printf("Caught! %i %f\n", i, cpl_table_get_double(rspec_temp, "WAVEL", i, NULL));
	    cpl_table_set_invalid(rspec_temp, "WAVEL", i);
        }
    }
    cpl_table_erase_invalid(rspec_temp);
    //espda_struct_check(rspec_temp, espda_struct_startend(rspec_temp, "WAVEL", 
    //    "WAVEL", wavel_start, wavel_end));

    // Compute a B-spline as final continuum -> not so effective
    //cpl_table *rspec_spline = espda_spec_rebin(rspec_rem,  3, 
    //wavel_start, wavel_end, 100);
    /*
    cpl_table *rspec_spline = cpl_table_duplicate(rspec_rem);
    const size_t n = cpl_table_get_nrow(rspec_spline);
    const size_t ncoeffs = (int)(n * 0.002); 
    const size_t k = 4;
    const size_t nbreak = ncoeffs + 2 - k;
    gsl_bspline_workspace *bw = gsl_bspline_alloc(k, nbreak); 
    gsl_vector *b = gsl_vector_alloc(ncoeffs);
    gsl_vector *x = gsl_vector_alloc(n);
    gsl_vector *y = gsl_vector_alloc(n);
    gsl_matrix *m = gsl_matrix_alloc(n, ncoeffs);
    gsl_vector *c = gsl_vector_alloc(ncoeffs);
    gsl_vector *w = gsl_vector_alloc(n);
    gsl_matrix *cov = gsl_matrix_alloc(ncoeffs, ncoeffs);
    gsl_multifit_linear_workspace *mw = gsl_multifit_linear_alloc(n, ncoeffs);
    for (int i = 0; i < n; i++) {
        gsl_vector_set(x, i, cpl_table_get_double(rspec_spline, "WAVEL", i, 
						  NULL));
        gsl_vector_set(y, i, cpl_table_get_double(rspec_spline, "FLUX", i, 
						  NULL));
        gsl_vector_set(w, i, 1 / 
		      (cpl_table_get_double(rspec_spline, "FLUXERR", i, NULL) *
		       cpl_table_get_double(rspec_spline, "FLUXERR", i, NULL)));
    }
    double kmin = cpl_table_get_double(rspec_spline, "WAVEL", 0, NULL); 
    double kmax = cpl_table_get_double(rspec_spline, "WAVEL", n - 1, NULL); 
    printf("%f %f\n", kmin, kmax);
    gsl_bspline_knots_uniform(kmin, kmax, bw); 
    for (int i = 0; i < n; i++) {
        double xi = gsl_vector_get(x, i);
	gsl_bspline_eval(xi, b, bw);
	for (int j = 0; j < ncoeffs; j++) {
	    double bj = gsl_vector_get(b, j);
	    gsl_matrix_set(m, i, j, bj);
	}
    }
    double bs_chisq;
    gsl_multifit_wlinear(m, w, y, c, cov, &bs_chisq, mw);
    for (int i = 0; i < n; i++) {
        double y_b, yerr_b;
	double x_b = cpl_table_get_double(rspec_spline, "WAVEL", i, NULL);
	gsl_bspline_eval(x_b, b, bw);
	gsl_multifit_linear_est(b, c, cov, &y_b, &yerr_b);
	cpl_table_set_double(rspec_spline, "FLUX", i, y_b);
	cpl_table_set_double(rspec_spline, "FLUXERR", i, yerr_b);
    }

    cpl_table_save(rspec_rem, NULL, NULL, "rspec_rem.fits", CPL_IO_CREATE);
    cpl_table_save(rspec_spline, NULL, NULL, "rspec_spline.fits", CPL_IO_CREATE);
    */

    // Compute a Savitzky-Golay filter as a local continuum
    /*
    cpl_table *rspec_sg = cpl_table_duplicate(rspec_rem);
    int sg_size = cpl_table_get_nrow(rspec_rem);
    double rem[sg_size];
    for (int i = 0; i < sg_size; i++) {
        rem[i] = cpl_table_get_double(rspec_rem, "FLUX", i, 0); 
    }  
    double sg[sg_size];
    char sg_out = sgfilter(rem, sg, sg_size, 400, 400, 0, 3);
    for (int i = 0; i < sg_size; i++) {
        cpl_table_set_double(rspec_sg, "CONT", i, sg[i]);
    }
    cpl_table_save(rspec_rem, NULL, NULL, "rspec_rem.fits", CPL_IO_CREATE);
    cpl_table_save(rspec_sg, NULL, NULL, "rspec_sg.fits", CPL_IO_CREATE);
    */
    //cpl_table *rspec_temp2 = espda_fit_sg(rspec_rem, "CONT", 401);
    //cpl_table_copy_data_double(rspec_temp, "CONT", 
    //cpl_table_get_data_double(rspec_temp2, "CONT"));

/*
    // Compute a running median as local continuum 
    cpl_table *rspec_rm = cpl_table_duplicate(rspec_rem);
    cpl_size rm_size = cpl_table_get_nrow(rspec_rem);
    cpl_size rm_hwind = 200;
#define MIN(X, Y) (((X) < (Y)) ? (X) : (Y))
#define MAX(X, Y) (((X) > (Y)) ? (X) : (Y))

    for (int i = 0; i < rm_size; i++) {
        cpl_size min = MAX(i - rm_hwind, 0); 
        cpl_size max = MIN(i + 1 + rm_hwind, rm_size);  
	cpl_table_unselect_all(rspec_rm);
	cpl_table_or_selected_window(rspec_rm, min, max-min);
        cpl_table *temp = cpl_table_extract_selected(rspec_rm);
	double med = cpl_table_get_column_median(temp, "FLUX");
	cpl_table_set_double(rspec_rm, "CONT", i, med);
    }  
    cpl_table_save(rspec_rm, NULL, NULL, "rspec_rm.fits", CPL_IO_CREATE);
*/

    cpl_table *fit  = NULL;
    espda_struct_copy_old(rspec_temp, "FIT", "FLUX");
    fit = cpl_table_duplicate(rspec_temp);
    //cpl_table_dump(fit, 0, 10, NULL);
    espda_util_safe(fspec_cont, espda_util_func(espda_fit_cspline(fit, "FIT", 
        fspec_cont, NULL), FIT_CSPLINE, 0));  
    espda_util_func(espda_fit_cspline(fit, "FIT", rspec_cont, NULL),
        FIT_CSPLINE, 0);
    cpl_table_delete(fit);

    cpl_table *cont = NULL;

    /*** UNCOMMENT HERE TO USE SAVITZKY-GOLAY ***/
    //rspec_temp = espda_fit_sg(rspec_rem, "CONT", sg_width_final);
    /***/
    /*
    //cpl_table_save(rspec_temp, NULL, NULL, "provona.fits", CPL_IO_CREATE);
    espda_struct_copy_old(rspec_temp, "FLUX_OLD", "FLUX");
    cpl_table_multiply_columns(rspec_temp, "CONT", "CONT_OLD");
    cpl_table_erase_column(rspec_temp, "FLUX_OLD");
    cpl_table_erase_column(rspec_temp, "CONT_OLD");
    //cpl_table_save(rspec_temp, NULL, NULL, "provetta.fits", CPL_IO_CREATE);
    */
    espda_struct_copy_old(rspec_temp, "CONT", "FLUX");


    cont = cpl_table_duplicate(rspec_temp);
    printf("%f, %f\n",
	   cpl_table_get_column_min(cont, "WAVEL"),
	   cpl_table_get_column_max(cont, "WAVEL"));
    if (fspec != NULL) {
        espda_util_func(espda_fit_cspline(cont, "CONT", fspec_cont, NULL), 
            FIT_CSPLINE, 0);  
        espda_util_func(espda_fit_cspline(cont, "CONT", fspec_rem, NULL),  
            FIT_CSPLINE, 0);  
    }
    espda_util_func(espda_fit_cspline(cont, "CONT", rspec_cont, NULL),
        FIT_CSPLINE, 0);
    printf("%f, %f\n",
	   cpl_table_get_column_min(rspec_cont, "WAVEL"),
	   cpl_table_get_column_max(rspec_cont, "WAVEL"));
    espda_util_func(espda_fit_cspline(cont, "CONT", rspec_rem, NULL),  
        FIT_CSPLINE, 0);  
    cpl_table_delete(cont);
    cpl_table_delete(rspec_cont_pre);
    cpl_table_delete(rspec_rem_pre);
    cpl_table_delete(rspec_temp);
    
    cpl_msg_info(cpl_func, "*** Extract final continuum ***");
    if (fspec != NULL) {
        espda_struct_norm(fspec_cont, "FLUX", "FLUXERR", "CONT", "NORM", "NORMERR");
        espda_struct_norm(fspec_rem,  "FLUX", "FLUXERR", "CONT", "NORM", "NORMERR");
    }
    espda_struct_norm(rspec_cont, "FLUX", "FLUXERR", "CONT", "NORM", "NORMERR");
    espda_struct_norm(rspec_rem,  "FLUX", "FLUXERR", "CONT", "NORM", "NORMERR");
    espda_struct_norm(fline_rem,  "PEAK", NULL,      "CONT", "NORM", NULL);

    cpl_table_copy_data_double(spline_cont, "CONT",
			       cpl_table_get_data_double(spline_cont, "FLUX"));
    
    //double chisq_line, rchisq_line = 0.0;
    //double chisq_cont, rchisq_cont = 0.0;
    double rchisq_line = 0.0;
    double rchisq_cont = 0.0;
    if (fspec != NULL) {
        double chisq_line = 0.0;
	double chisq_cont = 0.0;
        int dof = cpl_table_get_nrow(fspec_cont) - 
            cpl_table_get_nrow(fline_rem) * 3;
        espda_model_chisq(fspec_cont, "NORM", "NORMERR", "FIT",  dof, chisq_line, 
            rchisq_line);    
        espda_model_chisq(fspec_rem, "FLUX", "FLUXERR", "CONT", dof, chisq_cont, 
            rchisq_cont);
        cpl_msg_info(cpl_func, "Validation on the non-rebinned spectrum:");
        cpl_msg_info(cpl_func, " Reduced chi-squared on line fitting:       "
            "%7.4f (size: %4i, degrees of freedom: %4i)", rchisq_line, 
            (int)cpl_table_get_nrow(fspec_cont), dof);
        cpl_msg_info(cpl_func, " Reduced chi-squared on continuum fitting:  "
            "%7.4f (size: %4i, degrees of freedom: %4i)", rchisq_cont, 
            (int)cpl_table_get_nrow(fspec_rem), dof);
    }    
    espda_struct_sort(fline_rem, 1, "WAVEL", FALSE);
    
/*
 * Products are exported into frames.
 */ 

    cpl_msg_info(cpl_func, "*** Export products into frames ***");

    // Create propert lists
    cpl_propertylist *list = cpl_propertylist_new();
    cpl_propertylist *flist_cont = cpl_propertylist_duplicate(list);
    cpl_propertylist *flist_rem  = cpl_propertylist_duplicate(list);
    cpl_propertylist *rlist_cont = cpl_propertylist_duplicate(list);
    cpl_propertylist *rlist_rem  = cpl_propertylist_duplicate(list);
    cpl_propertylist *list_rem   = cpl_propertylist_duplicate(list);
    cpl_propertylist *spline_cont_list = cpl_propertylist_duplicate(list);
    if (!isinf(rchisq_cont) && !isnan(rchisq_cont)) {
        cpl_propertylist_append_double(flist_cont, QC_CONT_RCHISQ, rchisq_cont);
    }
    if (!isinf(rchisq_line) && !isnan(rchisq_line)) {
        cpl_propertylist_append_double(flist_rem,  QC_LINE_RCHISQ,  rchisq_line);
    }
    cpl_propertylist_append_double(list_rem,   DAS_Z_EM,        zem);
    cpl_propertylist_append_double(list_rem,   QC_COLDEN_THRES, colden_thres);
    cpl_propertylist_delete(list);
 
    // Create file names
    espda_frame_name(fname_cont, FSPEC_CONT);
    espda_frame_name(fname_rem,  FSPEC_REM);
    espda_frame_name(rname_cont, RSPEC_CONT);
    espda_frame_name(rname_rem,  RSPEC_REM);
    espda_frame_name(name_rem,   FLINE_REM);
    espda_frame_name(name_spline, "SPLINE_CONT");


    // Save frames
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fspec_cont, 
        flist_cont, "espda_fit_qsocont", FSPEC_CONT, fname_cont));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fspec_rem, 
        flist_rem,  "espda_fit_qsocont", FSPEC_REM,  fname_rem));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, rspec_cont, 
        rlist_cont, "espda_fit_qsocont", RSPEC_CONT, rname_cont));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, rspec_rem, 
        rlist_rem,  "espda_fit_qsocont", RSPEC_REM,  rname_rem));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, fline_rem,   
        list_rem,   "espda_fit_qsocont", FLINE_REM,  name_rem));
    espda_dfs_call(0, espda_frame_export(frameset, parlist, spline_cont,   
        spline_cont_list, "espda_fit_qsocont", "SPLINE_CONT", name_spline));

    cpl_msg_info(cpl_func, "*** Success!! ***");

    return 0;
}

/**@}*/
